"""The 'validateExistingFile' function validates that a given 'str' object
is a valid file or directory path that does not already exist. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.text import monoSpace

from worktoy.waitaminute import PathSyntaxException

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Union, TypeAlias, LiteralString

  Path: TypeAlias = Union[str, bytes, LiteralString]


def validateAvailablePath(path: Path) -> str:
  """
  Validates that a given 'str' object is a valid file or directory path
  that does not already exist.

  Args:
    path (str): The path to validate.

  Returns:
    str: The validated path.

  Raises:
    FileExistsError: If the file or directory already exists.
    NotADirectoryError: If the path is not a directory.
  """
  if not os.path.isabs(path):
    raise PathSyntaxException(path)
  if os.path.exists(path):
    infoSpec = """The path '%s' already exists!"""
    info = monoSpace(infoSpec % path)
    raise FileExistsError(info)
  return os.path.normpath(path)



"""HookException is raised from the AbstractNamespace class to wrap
exceptions raised by __getitem__ hooks. This is necessary to avoid
confusion with the expected KeyError exception in the metacall system."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.mcls import AbstractNamespace
  from worktoy.mcls.hooks import AbstractHook

  from typing import Any, Callable, TypeAlias, Self


class HookException(Exception):
  """
  This custom exception allows get item hooks to interrupt calls to
  __getitem__. Because the metacall system requires the __getitem__ to
  specifically raise a KeyError in certain situations, an exception raised
  by a hook might be confused for the KeyError. Instead,
  the AbstractNamespace class will catch exceptions raised by hooks and
  raise them from this exception:
  For example:

  try:
    hook(self, key, val)
  except Exception as exception:
    raise HookException(exception) from exception
  """

  initialException = _Attribute()
  namespaceObject = _Attribute()
  itemKey = _Attribute()
  errorValue = _Attribute()
  hookFunction = _Attribute()

  def __init__(
      self,
      exception: Exception,
      namespace: AbstractNamespace,
      key: str,
      val: object,
      hook: AbstractHook,
  ) -> None:
    self.initialException = exception
    self.namespaceObject = namespace
    self.itemKey = key
    self.errorValue = val
    self.hookFunction = hook
    Exception.__init__(self, str(exception))

  def _resolveOther(self, other: object) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the exception to another object."""
    cls = type(self)
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    if isinstance(other, cls):
      if self.initialException != other.initialException:
        return False
      if self.namespaceObject != other.namespaceObject:
        return False
      if self.itemKey != other.itemKey:
        return False
      if self.errorValue != other.errorValue:
        return False
      if self.hookFunction != other.hookFunction:
        return False
      return True
    return False



"""ReservedName is raised to indicate that a used name is reserved."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class ReservedName(Exception):
  """ReservedName is raised to indicate that a used name is reserved."""

  resName = _Attribute()

  def __init__(self, name: str) -> None:
    """Initialize the ReservedName object."""
    self.resName = name
    info = """Attempted to use reserved name: '%s'!"""
    Exception.__init__(self, info % name)

  def _resolveOther(self, other: Any) -> ReservedName:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the ReservedName object with another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    cls = type(self)
    if isinstance(other, cls):
      if self.resName != other.resName:
        return False
      return True
    return False



"""The 'worktoy.waitaminute' module provides custom exception classes. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from ._attribute import _Attribute
from ._attribute_error_factory import attributeErrorFactory
from ._zeroton_case_exception import ZerotonCaseException
from ._bad_value import BadValue
from ._bad_set import BadSet
from ._bad_delete import BadDelete
from ._metaclass_exception import MetaclassException
from ._alias_exception import AliasException
from ._abstract_exception import AbstractException
from ._duplicate_hook_error import DuplicateHookError
from ._hash_mismatch import HashMismatch
from ._cast_mismatch import CastMismatch
from ._flex_mismatch import FlexMismatch
from ._illegal_dispatcher import IllegalDispatcher
from ._hash_error import HashError
from ._unpack_exception import UnpackException
from ._dispatch_exception import DispatchException
from ._missing_variable import MissingVariable
from ._variable_not_none import VariableNotNone
from ._questionable_syntax import QuestionableSyntax
from ._read_only_error import ReadOnlyError
from ._reserved_name import ReservedName
from ._hook_exception import HookException
from ._write_once_error import WriteOnceError
from ._subclass_exception import SubclassException
from ._protected_error import ProtectedError
from ._type_exception import TypeException
from ._path_syntax_exception import PathSyntaxException
from ._illegal_instantiation import IllegalInstantiation
from ._cascade_exception import CascadeException
from ._del_exception import DelException

__all__ = [
    'attributeErrorFactory',
    'BadValue',
    'BadSet',
    'BadDelete',
    'MetaclassException',
    'AliasException',
    'HashMismatch',
    'CastMismatch',
    'FlexMismatch',
    'IllegalDispatcher',
    'InstanceException',
    'DuplicateHookError',
    'HashError',
    'UnpackException',
    'IllegalName',
    'DispatchException',
    'ParseException',
    'ResolveException',
    'MissingVariable',
    'THISContextException',
    'VariableNotNone',
    'QuestionableSyntax',
    'ReadOnlyError',
    'ReservedName',
    'HookException',
    'UnrecognizedMember',
    'WriteOnceError',
    'SubclassException',
    'ProtectedError',
    'TypeException',
    'PathSyntaxException',
    'IllegalInstantiation',
    'ZerotonCaseException',
    'CascadeException',
    'DelException'
]



"""DispatchException provides a custom exception raised when an instance
of OverloadDispatcher fails to resolve the correct function from the
given arguments. Because the overload protocol relies on type matching,
this exception subclasses TypeError such that it can be caught by external
error handlers. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.static import Dispatch

  from typing import Self


class DispatchException(TypeError):
  """DispatchException provides a custom exception raised when an instance
  of OverloadDispatcher fails to resolve the correct function from the
  given arguments. Because the overload protocol relies on type matching,
  this exception subclasses TypeError such that it can be caught by external
  error handlers. """

  dispatchObject = _Attribute()
  receivedArguments = _Attribute()
  receivedTypes = _Attribute()

  def __init__(self, dispatch: Dispatch, *args) -> None:
    self.dispatchObject = dispatch
    self.receivedArguments = args
    self.receivedTypes = [type(arg) for arg in args]

    ownerName = dispatch.getFieldOwner().__name__
    fieldName = dispatch.getFieldName()
    clsName = type(dispatch).__name__
    header = '%s object at %s.%s' % (clsName, ownerName, fieldName)
    typeArgs = [type(arg).__name__ for arg in args]
    argStr = '%s' % ', '.join(typeArgs)
    typeSigs = dispatch.getTypeSigs()
    typeSigStr = [str(sig) for sig in typeSigs]
    sigStr = '<br><tab>'.join(typeSigStr)

    infoSpec = """%s received arguments with signature: <br><tab>%s
    <br>which does not match any of the expected signatures:<br><tab>%s"""
    info = monoSpace(infoSpec % (header, argStr, sigStr))
    self.fuckUnitTest = info
    TypeError.__init__(self, info)

  def _resolveOther(self, other: object) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (list, tuple)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Check if two DispatchExceptions are equal."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    if self is other:
      return True
    if self.dispatchObject != other.dispatchObject:
      return False
    if self.receivedArguments != other.receivedArguments:
      return False
    if self.receivedTypes != other.receivedTypes:
      return False
    return True

  def __repr__(self, ) -> str:
    """
    Get the string representation of the DispatchException even if someone
    tries to repr it instead.
    """
    return self.__str__()



"""
The 'mostSpecificBase' function takes any number of classes and returns the
most specific class that is a base class of all the provided classes.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations


def mostSpecificBase(*classes: type, **kwargs) -> type:
  """
  The 'mostSpecificBase' function takes any number of classes and returns the
  most specific class that is a base class of all the provided classes.

  :param classes: A variable number of class types.
  :return: The most specific base class common to all provided classes.
  """
  if not classes:
    return object

  if len(classes) == 1:
    return classes[0]

  sample = kwargs.get('sample', None)
  if sample is None:
    sample = [b for b in classes[0].__mro__ if b is not object]
    prev = object
    return mostSpecificBase(*classes[1:], sample=sample, prev=prev)

  if not isinstance(sample, list):
    from worktoy.waitaminute import TypeException
    raise TypeException('sample', sample, list)

  prev = kwargs.get('prev', )
  if prev is None:
    from worktoy.waitaminute import TypeException
    raise TypeException('prev', prev, type)

  if not sample:
    return prev

  base = sample.pop(0)

  for cls in classes:
    if not issubclass(cls, base):
      return prev

  return mostSpecificBase(*classes, sample=sample, prev=base)



"""The 'worktoy.static' module provides low level parsing and casting
utilities. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

#  Private classes
from ._current_instance import _CurrentInstance, _InstanceAddress
from ._current_owner import _CurrentOwner, _OwnerAddress
from ._current_module import _CurrentModule
from ._current_class import _CurrentClass

#  Public classes
from ._item_call import ItemCall
from ._hist_dict import HistDict

#  Submodules
# from . import casting
from . import zeroton

#  Public classes
from ._abstract_object import AbstractObject  # Depends on zeroton

#  Private classes
from ._attribute import _Attribute  # Depends on AbstractObject
from ._alias import Alias
from ._pre_class import PreClass

#  Overloading classes
from ._type_sig import TypeSig
from ._dispatch import Dispatch
from ._overload import overload

__all__ = [
    'ItemCall',
    'HistDict',
    'AbstractObject',
    'Alias',
    'PreClass',
    'zeroton',
    'TypeSig',
    'Dispatch',
    'overload',
]



"""
The 'worktoy.static.zeroton' module provides the 'Zeroton' tokens.

Readers may be familiar with the term 'singleton' as a class restricted to
a single instance. This facilitates the design pattern where the class
provides stateless functionality whilst the instance holds an internal
state.

- Introducing the 'Zeroton'! -

It is just a 'singleton' as described above, but with zero instances,
hence the name. As such, the 'Zeroton' objects found here are entirely
stateless. 'Zeroton' itself is a metaclass facilitating the creation of
these zeroton objects.

- Motivation -
If Python did not already provide the 'super' function, it would have been
a prime candidate for implementation as a zeroton as it refers to an
object that cannot be resolved until runtime. The uses in this module are
similar except used to refer to objects that do not yet exist. As an
example, the 'typing.Self' object refers the current class scope, but with
the limitation that it is not really there. Instead, it is an abstraction
used in the type-hinting system. While this author holds those objects in
very high regard, as is evident across the codebase, 'worktoy' requires
an actual reference to the current scope.

- Introducing 'THIS' -
(Please note, that the 'THIS' object has no connection to the 'this'
module one can import to read about the 'this' statement in Python. The
nomenclature was committed to before this author ever learned about
'this'. The zeroton object is distinguished by being all uppercase.)

Python lacks function overloading as a language feature, but does provide
the most powerful abstraction in computer science: the metaclass. By
exposing nearly the entire class creation process, function overloading
and other advanced features are now only a metaclass away.
'worktoy.mcls.BaseMeta' is such a metaclass.

In the following example, 'ComplexNumber' derives from 'BaseMeta' and uses
function overloading to provide a very streamlined constructor system
supporting multiple type signatures. But what about a type signature
referring to 'ComplexNumber' itself? Since the overload decorator must be
present inside the class body. Resolving the super class is no trivial
matter, but it does exist, but in this case, a reference to an object not
yet existing is required. 'THIS' is that object. For example:


class ComplexNumber(metaclass=worktoy.mcls.BaseMeta):
  #  ComplexNumber defines its metaclass on the metaclass keyword. By
  #  default, this value is 'type', and unless a class uses a custom
  #  metaclass, 'type(SomeClass)' is always 'type'. Leading to the
  #  amusing fact: 'type(type) is type'.
  #  Please note the nomenclature: When a class uses a custom metaclass,
  #  it is said to 'be derived from' from that metaclass. To distinguish
  #  from 'normal' class inheritance, this author proposes that a subclass
  #  be said to 'be based on' the base class. The relationship between a
  #  subclass and a baseclass is entirely different from between a class
  #  and its metaclass.

  __slots__ = ('realPart', 'imagPart')

  @overload(float, float)  # Decorates the float-float constructor
  def __init__(self, x: float, y: float) -> None:
    self.realPart = x
    self.imagPart = y

  @overload(complex)  # Decorates the complex constructor
  def __init__(self, z: complex) -> None:
    self.__init__(z.real, z.imag)  # Recursively calls the float-float

  #  And now the type signature referring to the class under construction:
  @overload(THIS)  # Decorates the 'ComplexNumber' constructor
  def __init__(self, other: Self) -> None:
    #  Note the type-hinting to 'typing.Self'. 'THIS' implements the same
    #  but at runtime. As such, when instantiating 'ComplexNumber' by
    #  passing another instance of it, the function overloading system
    #  uses 'THIS' as a temporary token for the class that will be
    #  available later.
    self.__init__(other.realPart, other.imagPart)
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

#  Private list of disallowed names
from ._reserved_names import _reservedNames

#  The Zeroton namespace
from ._zero_space import ZeroSpace

#  The Zeroton metaclass
from ._zeroton import Zeroton

#  The objects derived from Zeroton
from ._deleted import DELETED
from ._this import THIS
from ._owner import OWNER
from ._desc import DESC

__all__ = [
    'Zeroton',
    'THIS',
    'OWNER',
    'DELETED',
    'DESC',
]



"""
The 'bipartiteMatching' function solves the bipartite matching problem by
recursive backtracking with forward checking returns the first encountered
valid mapping.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, TypeAlias, Union, Iterator

  SlotInts: TypeAlias = dict[Any, tuple[int, ...]]
  SlotMap: TypeAlias = dict[Any, int]


def bipartiteMatching(slots: list[tuple[int, ...]]) -> list[int]:
  """
  Recursive backtracking with forward checking. Given a list of tuples
  of candidate indices, one per slot, finds an assignment of one unique
  index to each slot, with no repeats. Returns list of assigned indices,
  or raises ValueError if no solution exists.
  """
  # Base case: success
  if not slots:
    return []

  # Step 1: fail if any empty
  for opts in slots:
    if not opts:
      raise ValueError('No valid assignment for slot')

  # Step 2: pick slot with fewest options
  idx, opts = min(enumerate(slots), key=lambda x: len(x[1]))

  # Step 3: try each option
  for value in opts:
    # Build reduced slots
    reduced = []
    for j, otherOpts in enumerate(slots):
      if j == idx:
        continue
      reducedOpts = tuple(v for v in otherOpts if v != value)
      reduced.append(reducedOpts)
    try:
      result = bipartiteMatching(reduced)
    except ValueError:
      continue
    # Rebuild full result
    result = result[:idx] + [value] + result[idx:]
    return result

  raise ValueError('No valid assignment for slot %d' % idx)



"""
CastMismatch is a custom exception raised to indicate that a 'TypeSig'
object has failed to cast a collection of arguments to itself.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import HashMismatch

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


class CastMismatch(HashMismatch):
  """
  CastMismatch is a custom exception raised to indicate that a 'TypeSig'
  object has failed to cast a collection of arguments to itself.
  """

  pass



"""
Cast provides casting from any number of positional arguments to one
target type.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from worktoy.static import AbstractObject, _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import TypeAlias, Callable, Any


class Cast(AbstractObject):
  """Cast provides casting from any number of positional arguments to one
  target type."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __target_type__ = None

  #  Public Variables
  target = _Attribute()  # No default target

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args, ) -> None:
    """
    Initialize the Cast instance with a target type.
    """
    for arg in args:
      if isinstance(arg, type):
        self.__target_type__ = arg
        break
    else:
      raise NotImplementedError('missing argument for target type')

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _instantiateTarget(self, *args, ) -> Any:  # target
    """
    Instantiate the target type with the given arguments.
    """
    return self.target(*args, )



"""
AbstractMetaclass provides the baseclass for custom metaclasses.
"""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType as Func

from ..static import HistDict
from ..waitaminute import QuestionableSyntax, TypeException, MissingVariable
from . import Base
from . import AbstractNamespace as ASpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any
else:
  pass


class _MetaMetaclass(type):
  """
  _MetaMetaclass is necessary for metaclasses to implement customized
  dunder methods such as __str__ and __repr__. Please note the pattern
  that metaclasses should have the meta-metaclass as both a b baseclass
  *and* metaclass. This is consistent with the default behaviour
  inheriting from 'type' whilst, possibly implicitly, using 'type' as the
  metaclass.
  """

  def __str__(cls, ) -> str:
    """Returns the name of the class. """
    return """%s[metaclass=%s]""" % (cls.__name__, cls.__class__.__name__)


class AbstractMetaclass(_MetaMetaclass, metaclass=_MetaMetaclass):
  """
  Abstract base for custom metaclasses that separates concerns between
  class construction and class behavior.

  This design delegates the initial class namespace to a custom object
  returned by '__prepare__', while keeping class semantics within the
  metaclass itself.

  The namespace object may define a method called 'compile()', which
  should return the finalized dictionary to be passed to 'type.__new__'.

  This separation enables:
  - Customizing how the class body is assembled without affecting the
    resulting class behavior.
  - Modifying what it means to be a class, independently of how the
    class body is constructed.

  - CUSTOM NAMESPACE COMPILATION -

  This module also provides an 'AbstractNamespace' class intended to be
  used with '__prepare__'. It defines a 'compile()' method and an
  '__init__()' signature compatible with the arguments passed to
  '__prepare__'.

  from . import AbstractNamespace as ASpace  # For brevity

  Specifically, '__prepare__' is defined as:
    def __prepare__(mcls, name: str, bases: Base, **kwargs) -> ASpace

  And 'ASpace' ('AbstractNamespace') implements:
    def __init__(self, mcls: type, name: str, bases: Base, **kwargs)

  This allows the namespace object to be instantiated with full context
  about the class being defined, while remaining isolated from the
  behavior of the resulting class itself. For more information about the
  class body execution flow, see the 'AbstractNamespace' class
  documentation.

  - CUSTOM CLASS CREATION -

  After assembly and compilation of the namespace, the metaclass itself
  takes responsibility for validating and finalizing the class.

  This includes a pass over the namespace via the '_validateNamespace'
  static method. It performs checks for common mistakes such as typos
  in special method names—for example:
    - '__set_item__' instead of '__setitem__'
    - '__get_attr__' instead of '__getattr__'
    - '__setname__' instead of '__set_name__'

  If any such names are found, a 'QuestionableSyntax' error is raised
  to prompt correction.

    -  '__del__' or '__delete__' ? -
  The descriptor protocol allows classes to define what happens when an
  attribute is deleted from an instance. This is handled by the '__delete__'
  method. It is much less common than '__get__' and '__set__', which govern
  attribute access and assignment, respectively.

  Because of the naming similarity to '__del__'—a special method for object
  finalization—it's easy to accidentally implement '__del__' when one meant
  '__delete__'.

  Bugs caused by incorrect use of `__del__`—especially when accidentally
  used instead of `__delete__`—are notoriously difficult to trace. Since
  `__del__` is called only when the object is garbage collected (which may
  be delayed or never happen), the consequences of the mistake are often
  deferred until long after the original action that should have triggered
  cleanup. This makes it extremely hard to correlate the broken behavior
  with its source. Worse still, because `__del__` doesn’t raise errors if
  used in the wrong context, failures are often silent, leading to
  inconsistent state, memory leaks, or subtle bugs in object lifecycles that
  resist even thorough debugging.

  For the above reasons, the 'worktoy' library will raise 'SyntaxError'
  whenever '__del__' is found in the namespace. If an implementation of
  '__del__' is actually intended, the class creation must be invoked
  with the keyword argument 'trustMeBro=True'.

    - Standard Methods -
  While not implemented in this metaclass, the following pattern allows
  sub-metaclasses to implement automatically generated methods in certain
  cases. The 'worktoy.ezdata.EZMeta' dataclass implementation exemplifies
  this pattern. Derived classes can define properties directly in the
  class body, and 'EZMeta' will automatically generate the necessary
  methods, such as '__init__', allowing for instantiation with either
  positional, keyword or even mixed arguments.

  The sub-metaclass would implement factories as static methods:

  from types import FunctionType as Func

  Bases: TypeAlias = tuple[type, ...]
  Space: TypeAlias = dict[str, Any]

  @staticmethod
  def __init_factory__(name: str, bases: Bases, space: Space, **kw) -> Func

  @staticmethod
  def __new_factory__(name: str, bases: Bases, space: Space, **kw) -> Func

  @staticmethod
  def __str_factory__(name: str, bases: Bases, space: Space, **kw) -> Func

  Finally, the sub-metaclass would have to implement:

  @staticmethod
  def autoGenMethods(name: str, bases: Bases, space: Space, **kw) -> Space:
    This method would be invoked after the namespace has been validated
    and before the class is created. It would return a modified namespace
    with the necessary methods added.

    - Notifying Baseclasses -
  Having validated the namespace, the metaclass falls back to type.__new__
  and returns the newly created class. Finally, this class arrives in the
  '__init__' method where the metaclass notifies any baseclass that
  implements the '__subclasshook__' method of the class creation. This
  marks the end of the class creation process.

  - CUSTOM CLASS BEHAVIOR -

  Once a class has been created by the metaclass system, it may define
  its own runtime behavior by implementing special methods prefixed with
  `__class_`. These allow the class object itself to participate directly
  in common operations such as being called, iterated, or printed.

  These methods resemble `__class_getitem__` from standard Python but are
  more general. Each of them overrides a specific class-level behavior.

  The following hooks relate to common class-level operations:

  - __class_call__(cls, *args, **kwargs) -> Any
    Called when the class object is called like a function. Overrides the
    default behavior of constructing instances. Can be used to implement
    singletons, factories, registries, etc.

  - __class_instancecheck__(cls, obj: Any) -> bool
    Called during isinstance(obj, cls). Controls how instance membership
    is determined. Supersedes metaclass-level __instancecheck__.

  - __class_subclasscheck__(cls, sub: type) -> bool
    Called during issubclass(sub, cls). Controls dynamic subclass logic.
    Allows behavior similar to abstract base classes or trait systems.

  #  The following hooks allow classes to define how they are printed

  - __class_str__(cls) -> str
    Called when str(cls) is invoked. Provides human-readable string form
    for dynamically generated or aliased classes.

  - __class_repr__(cls) -> str
    Called when repr(cls) is invoked. Allows classes to override their
    debug representation.

  #  The following hooks relate to class-level iteration

  - __class_iter__(cls) -> Iterator
    Called when iter(cls) is invoked. Makes the class object iterable.
    Useful for registry-style classes, enums, and similar patterns.

  - __class_next__(cls) -> Any
    Called when next(cls) is invoked. Meaningful only if the class itself
    is its own iterator as returned by __class_iter__.

  - __class_bool__(cls) -> bool
    Called when bool(cls) is invoked. Allows classes to define their truth
    value. By default, every class is 'truthy'.

  - __class_contains__(cls, item: Any) -> bool
    Allows classes to define membership checks on the class level. By
    default, this checks if the item is an instance of the class itself.

  - __class_len__(cls) -> int
    Called when len(cls) is invoked.

  - __class_hash__(cls) -> int
    Called when hash(cls) is invoked. Allows classes to define their own
    hash value. Defaults to:
    mcls = type(cls)  # The metaclass of the class
    baseNames = [b.__name__ for b in cls.__bases__]
    return hash((cls.__name__, *baseNames, mcls.__name__))
    #  PLEASE NOTE: The 'overload' protocol provided by the 'worktoy'
    library expects this exact hash value. Reimplementing the hash value
    will make the dispatching of overloads unable to 'fast' recognize the
    class.

  - __class_eq__(cls, other: Any) -> bool
    Called to allow classes to equal each other. Please note that this
    inclusion is for completeness more than anything else. The '__eq__' in
    this metaclass does look for '__class_eq__' on the class, but falls
    back to __class_hash__.

  The following hooks allows dictionary-like access to the class.

  - __class_getitem__(cls, item: Any) -> Any
    Called when cls[item] is invoked. Please note that this is already
    implemented in Python 3.7+ as a standard class method. It is listed
    here only for completeness. This means that this metaclass does not
    need to implement '__getitem__' to look for the '__class_getitem__' on
    the class itself. In fact, the __getitem__ on the metaclass would only
    ever be invoked if Foo['bar'] is invoked on a class Foo that does not
    implement '__class_getitem__'.

  - __class_setitem__(cls, item: Any, value: Any) -> None
    Called when cls[item] = value is invoked. 
    
  - __class_delitem__(cls, item: Any) -> None
    Called when del cls[item] is invoked.

    - Class Attribute Hooks -

  - __class_getattr__(name: str, exception: Exception) -> Any
    If a non-existing attribute is attempted accessed on a class object,
    the '__getattr__' method on the metaclass is invoked. This method
    allows the class itself to handle this case. It is strongly advised
    that this method, and '__getattr__' in general, raises an
    AttributeError unless the key passed to it has a valid and sensible
    meaning in the context.

  The following hooks allow classes to define custom behavior for
  attribute assignment and deletion at the class level.

  - __class_setattr__(name: str, value: Any) -> None
  - __class_delattr__(name: str) -> None

  The following hooks would be relevant only for nested classes that
  implement the descriptor protocol. These class-level descriptor hooks
  remain unimplemented due to unresolved hazards in Python's class
  construction behavior. In particular, referencing other class objects
  while a metaclass is "awake" (i.e., inside its __prepare__, __new__,
  or __init__) can lead to context leakage. Python may interpret unrelated
  class references within the scope of the active metaclass, sometimes
  routing calls to the wrong metaclass entirely.

  - __class_get__(cls, instance: Any, owner: type) -> Any
  - __class_set__(cls, instance: Any, value: Any) -> None
  - __class_delete__(cls, instance: Any) -> None
  - __class_set_name__(cls, owner: type, name: str) -> None

  Finally, the following hooks are logically meaningless:

  - __class_init__(...)
    There is no sensible point at which this would be called. Class
    objects are constructed by metaclasses, and no outer context exists
    in which to simulate an “init” phase.

  - __class_new__(...)
    Similar to above, class creation is driven by the metaclass’s
    __new__, and there is no standard mechanism for a class to override
    or participate in that step on its own behalf.

  - __class_del__(...)
    This remains unimplemented for the same reason as why the namespace
    validator described above raises a SyntaxError when it encounters
    '__del__' in the namespace.

  - __class_getattribute__(...)
    [REDACTED: Cognito Hazard]
  """

  @classmethod
  def __prepare__(mcls, name: str, bases: Base, **kwargs) -> ASpace:
    """The __prepare__ method is invoked before the class is created. This
    implementation ensures that the created class has access to the safe
    __init__ and __init_subclass__ through the BaseObject class in its
    method resolution order."""
    return ASpace(mcls, name, bases, **kwargs)

  def __new__(mcls, name: str, bases: Base, space: ASpace, **kw) -> type:
    """The __new__ method is invoked to create the class."""
    namespace = mcls._validateNamespace(name, bases, space.compile(), **kw)
    return _MetaMetaclass.__new__(mcls, name, bases, namespace, **kw)

  def __init__(cls, name: str, bases: Base, space: ASpace, **kwargs) -> None:
    """The __init__ method is invoked to initialize the class."""
    if TYPE_CHECKING:
      assert isinstance(space, ASpace)
      assert isinstance(bases, tuple)
    _MetaMetaclass.__init__(cls, name, bases, space, **kwargs)
    cls._notifySubclassHook(cls, *bases)

  def __call__(cls, *args, **kwargs) -> Any:
    """The __call__ method is invoked when the class is called."""
    try:
      func = cls._dispatchClassHook('__class_call__')
    except NotImplementedError:
      return super().__call__(*args, **kwargs)  # NOQA
    else:
      return func(cls, *args, **kwargs)

  def __instancecheck__(cls, instance) -> bool:
    """
    This implementation allows the class to customize the instance
    check by implementing a method called __class_instancecheck__.
    """
    try:
      func = cls._dispatchClassHook('__class_instancecheck__')
    except NotImplementedError:
      return super().__instancecheck__(instance)  # NOQA
    else:
      return True if func(cls, instance) else False

  def __subclasscheck__(cls, subclass) -> bool:
    """
    This implementation allows the class to customize the subclass
    check by implementing a method called __class_subclasscheck__.
    """
    try:
      func = cls._dispatchClassHook('__class_subclasscheck__')
    except NotImplementedError:
      return super().__subclasscheck__(subclass)  # NOQA
    else:
      return True if func(cls, subclass) else False

  def __str__(cls, ) -> str:
    """The __str__ method is invoked to get the string representation of
    the class."""
    try:
      func = cls._dispatchClassHook('__class_str__')
    except NotImplementedError:
      return super().__str__()
    else:
      return func(cls)

  def __repr__(cls, ) -> str:
    """The __repr__ method is invoked to get the string representation of
    the class."""
    try:
      func = cls._dispatchClassHook('__class_repr__')
    except NotImplementedError:
      return super().__repr__()
    else:
      return func(cls)

  def __iter__(cls, ) -> Any:
    """
    The __iter__ method is invoked to iterate over the class.
    """
    try:
      func = cls._dispatchClassHook('__class_iter__')
    except NotImplementedError:
      try:
        out = super().__iter__()  # NOQA
      except AttributeError as attributeError:
        if '__iter__' in str(attributeError):
          infoSpec = """'%s' object is not iterable"""
          info = infoSpec % cls.__name__
          raise TypeError(info) from attributeError
      except TypeError as typeError:
        raise
    else:
      return func(cls)

  def __next__(cls, ) -> Any:
    """The __next__ method is invoked to get the next item in the class."""
    try:
      func = cls._dispatchClassHook('__class_next__')
    except NotImplementedError:
      return super().__next__()  # NOQA
    else:
      return func(cls)

  def __bool__(cls, ) -> bool:
    """The __bool__ method is invoked to get the truth value of the
    class."""
    try:
      func = cls._dispatchClassHook('__class_bool__')
    except NotImplementedError:
      return True  # By default, classes are truthy
    else:
      return True if func(cls) else False

  def __contains__(cls, other: Any) -> bool:
    """The __contains__ method is invoked to check if the item is in the
    class."""
    try:
      containsFunc = cls._dispatchClassHook('__class_contains__')
    except NotImplementedError:
      pass
    else:
      return True if containsFunc(cls, other) else False
    try:
      for item in cls:
        if item == other:
          break
      else:
        return False
    except Exception as exception:
      raise
    else:
      return True

  def __len__(cls, ) -> int:
    """The __len__ method is invoked to get the length of the class."""
    lenFunc, value = None, None
    try:
      lenFunc = cls._dispatchClassHook('__class_len__')
    except NotImplementedError:
      pass
    else:
      return lenFunc(cls)
    try:
      value = 0
      for _ in cls:
        value += 1
    except Exception as exception:
      if 'object is not iterable' in str(exception):
        infoSpec = """'%s' object has no len()"""
        info = infoSpec % cls.__name__
        raise TypeError(info) from exception
      raise
    else:
      return value

  def __hash__(cls, ) -> int:
    """The __hash__ method is invoked to get the hash value of the class.
    This is used to identify the class in dictionaries and sets."""
    try:
      func = cls._dispatchClassHook('__class_hash__')
    except NotImplementedError:
      baseNames = [b.__name__ for b in cls.__bases__]
      mcls = type(cls)
      return hash((cls.__name__, *baseNames, mcls.__name__))
    else:
      return func(cls)

  def __eq__(cls, other: Any) -> bool:
    """
    The __eq__ method is invoked to check if the class is equal to
    another object.
    """
    if cls is other:
      return True
    try:
      func = cls._dispatchClassHook('__class_eq__')
    except NotImplementedError:
      pass
    else:
      return True if func(cls, other) else False
    try:
      hashFunc = cls._dispatchClassHook('__class_hash__')
    except NotImplementedError:
      return False
    else:
      return True if hashFunc(cls) == hash(other) else False

  def __getattr__(cls, key: str) -> Any:
    """
    The __getattr__ method is invoked to get the attribute of the
    class.
    """
    try:
      func = cls._dispatchClassHook('__class_getattr__', )
    except NotImplementedError:
      infoSpec = """type object '%s' has no attribute '%s'"""
      info = infoSpec % (cls.__name__, key)
      raise AttributeError(info)
    else:
      return func(cls, key)

  def __setattr__(cls, name: str, value: Any) -> None:
    """The __setattr__ method is invoked to set the attribute of the
    class."""
    try:
      func = cls._dispatchClassHook('__class_setattr__')
    except NotImplementedError:
      return super().__setattr__(name, value)
    else:
      return func(cls, name, value)

  def __delattr__(cls, name: str) -> None:
    """The __delattr__ method is invoked to delete the attribute of the
    class."""
    try:
      func = cls._dispatchClassHook('__class_delattr__')
    except NotImplementedError:
      return super().__delattr__(name)
    else:
      return func(cls, name)

  #  Notice the absense of __getitem__ here. If implemented, it would
  #  override the standard Python behavior.

  def __setitem__(cls, item: Any, value: Any) -> None:
    """The __setitem__ method is invoked to set the item in the class."""
    try:
      func = cls._dispatchClassHook('__class_setitem__')
    except NotImplementedError:
      return super().__setitem__(item, value)  # NOQA
    else:
      return func(cls, item, value)

  def __delitem__(cls, item: Any) -> None:
    """The __delitem__ method is invoked to delete the item from the
    class."""
    try:
      func = cls._dispatchClassHook('__class_delitem__')
    except NotImplementedError:
      return super().__delitem__(item)  # NOQA
    else:
      return func(cls, item)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _validateNamespace(
      mcls, name: str,
      bases: Base,
      space: dict,
      **kwargs,
  ) -> dict:
    """
    The _validateNamespace method is invoked to validate the namespace
    object before the class is created.
    """
    if '__del__' in space and '__delete__' not in space:
      if not kwargs.get('trustMeBro', False):
        raise DelException(mcls, name, bases, space)
    derpNames = [
        '__get_item__', '__set_item__', '__get_attr__',
        '__set_attr__'
    ]
    realNames = ['__getitem__', '__setitem__', '__getattr__', '__setattr__']
    for derp, name in zip(derpNames, realNames):
      if derp in space:
        raise QuestionableSyntax(derp, name, )
    return space

  @staticmethod
  def _notifySubclassHook(cls, *bases) -> type:
    """The _notifySubclassHook method is invoked to notify each baseclass
    of the created class of the class creation."""
    for base in bases:
      hook = getattr(base, '__subclasshook__', None)
      if hook is None:
        continue
      hook(cls)
    return cls

  def _dispatchClassHook(cls, name: str, ) -> Func:
    """
    If the class implements a method called '__class_[name]__', this
    method returns the underlying function object. Otherwise, it raises
    'NotImplementedError'.
    """
    try:
      func = object.__getattribute__(cls, name)
    except AttributeError as attributeError:
      raise NotImplementedError from attributeError
    else:
      if hasattr(func, '__func__'):
        return func.__func__
      if isinstance(func, Func):
        return func
      raise TypeException(name, func, Func)
    finally:
      if TYPE_CHECKING:
        pycharmPlease = 69420
        assert isinstance(pycharmPlease, Func)
        return pycharmPlease
      else:
        pass

  def getNamespace(cls) -> ASpace:
    """Get the namespace object for the class."""
    if TYPE_CHECKING:
      assert isinstance(cls, AbstractMetaclass)
    space = getattr(cls, '__namespace__', None)
    if space is None:
      raise MissingVariable('__namespace__', ASpace, HistDict, dict)
    if isinstance(space, dict):
      if TYPE_CHECKING:
        assert isinstance(space, ASpace)
      return space
    raise TypeException('__namespace__', space, ASpace, HistDict, dict)



"""IllegalInstantiation is a custom exception raised to indicate than
an attempt was made to instantiate a class under illegal conditions."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Never


class IllegalInstantiation(TypeError):
  """IllegalInstantiation is a custom exception raised to indicate than
  an attempt was made to instantiate a class under illegal conditions."""

  cls = _Attribute()

  def __init__(self, cls_: type) -> None:
    """Initialize the IllegalInstantiation with the class."""
    self.cls = cls_
    infoSpec = """Illegal instantiation of class '%s'"""
    info = infoSpec % cls_.__name__
    TypeError.__init__(self, info)



"""
'DESC' provides a reference to a descriptor in a context where the actual
descriptor class to be referenced does not yet exist.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...text import monoSpace

from . import Zeroton


class DESC(metaclass=Zeroton):
  """DESC is a token object indicating the descriptor object of the
  surrounding scope. """

  @classmethod
  def __class_str__(cls, ) -> str:
    """Required explanation of the class. """
    infoSpec = """The '%s' Zeroton is a placeholder for the descriptor 
    object of the surrounding scope. """
    return monoSpace(infoSpec % cls.__name__)



"""The 'maybe' function returns the first received argument that is not
None. Thus it provides a None-aware filter. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


def maybe(*args: Any) -> Any:
  """The 'maybe' function returns the first received argument that is not
  None. Thus, it provides a None-aware filter.

  Args:
      *args: The arguments to filter.

  Returns:
      The first argument that is not None.
  """
  for arg in args:
    if arg is not None:
      return arg


def maybeType(type_: type, *args) -> Any:
  """The 'maybeType' function returns the first received argument that is not
  None and is of the specified type. Thus, it provides a None-aware filter
  with type checking.

  Args:
      type_: The type to check against.
      *args: The arguments to filter.

  Returns:
      The first argument that is not None and is of the specified type.
  """
  for arg in args:
    if isinstance(arg, type_):
      return arg


def maybeTypes(type_: type, *args) -> Any:
  """The 'maybeTypes' function returns the first received argument that is
  not
  None and is of the specified type. Thus, it provides a None-aware filter
  with type checking.

  Args:
      type_: The type to check against.
      *args: The arguments to filter.

  Returns:
      The first argument that is not None and is of the specified type.
  """
  typeArgs = []
  for arg in args:
    if isinstance(arg, type_):
      typeArgs.append(arg)
  return (*typeArgs,)



"""
DelException is a custom exception raised when someone attempts to create
a class that implements the '__del__' method without providing the custom
keyword argument: 'trustMeBro=True'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Never


class DelException(SyntaxError):
  """
  DelException is a custom exception raised when someone attempts to create
  a class that implements the '__del__' method without providing the custom
  keyword argument: 'trustMeBro=True'.
  """

  mcls: _Attribute()
  name: _Attribute()
  bases: _Attribute()

  def __init__(self, *args) -> None:
    """Initialize the DelException with the class."""
    _mcls, _name, _bases = None, None, None

    for arg in args:
      if isinstance(arg, type) and _mcls is None:
        _mcls = arg
        continue
      if isinstance(arg, str) and _name is None:
        _name = arg
        continue
      if isinstance(arg, tuple) and _bases is None:
        for base in arg:
          if not isinstance(base, type):
            break
        else:
          _bases = arg
          continue
      if all(i is not None for i in (_mcls, _name, _bases)):
        break

    if _mcls is not None:
      self.mcls = _mcls
    if _name is not None:
      self.name = _name
    if _bases is not None:
      self.bases = _bases

    SyntaxError.__init__(self, self.__str__())

  def __str__(self) -> str:
    """Return a string representation of the DelException."""
    infoSpec = """When attempting to derive a class named '%s' from the 
    metaclass '%s', the '__del__' method was found in the namespace! This 
    is almost always a typo, but if not this error can be suppressed by 
    passing the keyword argument 'trustMeBro=True' during class creation. """
    if self.bases:
      mclsSpec = """%s with bases: (%s)"""
    else:
      mclsSpec = """%s%s"""
    basesStr = ', '.join(base.__name__ for base in self.bases)
    mclsName = mclsSpec % (self.mcls.__name__, basesStr)
    info = infoSpec % (self.name, mclsName)
    return info



"""Cast creates an AbstractCast subclass object appropriate to a given
type. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from . import AbstractCast, AutoCast, IntCast
from . import FloatCast, ComplexCast

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


class Cast(AbstractCast):
  """Cast creates an AbstractCast subclass object appropriate to a given
  type. """

  def getTargetType(self, **kwargs) -> type:
    """This subclass should not be used directly. """
    e = """This subclass should not be used directly. """
    raise RuntimeError(e)

  def _cast(self, *args, **kwargs) -> Any:
    """This subclass should not be used directly. """
    e = """This subclass should not be used directly. """
    raise RuntimeError(e)

  def __new__(cls, targetType: type) -> AbstractCast:
    """Create a new Cast object."""
    if targetType is int:
      return IntCast()
    if targetType is float:
      return FloatCast()
    if targetType is complex:
      return ComplexCast()
    return AutoCast(targetType)



"""
NameHook filters named used in the namespace system.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...waitaminute import QuestionableSyntax

from . import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, TypeAlias

  NearMiss: TypeAlias = tuple[str, str]


class NameHook(AbstractHook):
  """
  NameHook intercepts names added to the namespace and filters out
  "near-miss" identifiers that resemble critical Python dunder methods.
  These mistakes often go unnoticed, leading to subtle bugs or broken
  protocol support.

  This hook raises a QuestionableSyntax exception when such names are
  detected during assignment in the namespace.

  ## Purpose

  Many magic methods in Python have specific names that must be spelled
  exactly. If a user misspells one by inserting or omitting underscores,
  the name is silently ignored by Python and treated as an ordinary
  attribute — sometimes shadowing a builtin or behaving unexpectedly.

  ## Near-miss Examples
  ___________________________________________________________________________
  | Intended Name  | Mistyped Name | Notes                                 |
  |----------------|---------------|---------------------------------------|
  | `__set_name__` | `__setname__` | Misses descriptor registration        |
  | `__getitem__`  | `__get_item__`| Breaks item access in dict-like APIs  |
  | `__setitem__`  | `__set_item__`| Same as above                         |
  | `__delitem__`  | `__del_item__`| Silent failure of delete protocol     |
  | `__delete__`   | `__del__`     | High risk: __del__ ties to GC hooks   |
  ¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨
  These errors can be especially difficult to diagnose, as they usually
  do not raise any errors directly — instead, they silently fail to
  participate in expected behaviors or override builtin methods.

  ## Usage
  To use NameHook, simply declare it in your namespace class:

  class Space(AbstractNamespace):  # Must inherit from AbstractNamespace
    #  Custom namespace class inheriting from AbstractNamespace
    nameHook = NameHook()  # Register the hook
"""

  @classmethod
  def _getNearMisses(cls) -> list[NearMiss]:
    """
    Get the near-miss names.
    """
    return [
        ('__set_name__', '__setname__'),  # NOQA, miss-spelled name
        ('__getitem__', '__get_item__'),
        ('__setitem__', '__set_item__'),
        ('__delitem__', '__del_item__'),
        ('__delete__', '__del__'),
    ]

  @classmethod
  def _validateName(cls, name: str) -> bool:
    """
    Compares the name to list of potential near-miss names. If the name
    is a near-miss, a QuestionableSyntax exception is raised.
    """
    nearMisses = cls._getNearMisses()
    for nearMiss in nearMisses:
      if name == nearMiss[1]:
        raise QuestionableSyntax(*nearMiss, )
    return False

  def setItemHook(self, key: str, value: Any, oldValue: Any) -> bool:
    """
    Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False.
    """
    return self._validateName(key)



"""VariableNotNone should be raised when a variable is unexpectedly not
None."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import monoSpace

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self


class VariableNotNone(Exception):
  """VariableNotNone should be raised when a variable is unexpectedly not
  None."""

  name = _Attribute()
  value = _Attribute()

  def __init__(self, *args) -> None:
    """Initialize the VariableNotNone object."""
    _name, _value = None, None
    for arg in args:
      if isinstance(arg, str):
        if _name is None:
          _name = arg
          continue
      if _value is None:
        _value = arg
        break
    if _name is not None:
      self.name = _name
    if _value is not None:
      self.value = _value
    Exception.__init__(self, )

  def __str__(self, ) -> str:
    """Get the info spec."""
    if self.name is None and self.value is None:
      infoSpec = """Unexpected value at name expected to be None!%s%s"""
      name, valueStr = '', ''
    elif self.name is None:
      infoSpec = """Unexpected value: '%s' at name expected to be 
      None!%s"""
      name = ''
      valueStr = monoSpace(str(self.value), )
    elif self.value is None:
      infoSpec = """Unexpected value at name '%s' expected to be 
      None!"""
      name = self.name
      valueStr = ''
    else:
      infoSpec = """Unexpected value: '%s' at name '%s' expected to be 
      None!"""
      valueStr = monoSpace(str(self.value), )
      name = self.name
    return monoSpace(infoSpec % (valueStr, name), )

  __repr__ = __str__



"""
KeeHook provides the namespace hook for the KeeSpace namespace class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static import AbstractObject
from ..mcls.hooks import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any

  from . import _AutoMember


class KeeHook(AbstractHook):
  """
  KeeHook provides the namespace hook for the KeeSpace namespace class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _addMember(self, name: str, value: Any, **kwargs) -> None:
    """
    Add a member to the KeeSpace namespace.
    """
    if value is None:
      if kwargs.get('_recursion2', False):
        raise RecursionError
      return self._addMember(name, name, _recursion2=True)
    valueType = self._getValueType()
    if valueType is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self._setValueType(type(value))
      return self._addMember(name, value, _recursion=True)
    else:
      if not isinstance(value, valueType):
        es = """Found inconsistent value type for members of the '%s' class 
        under construction. Member named: '%s' has value type: '%s', 
        while the previous values had value type: '%s'. Values passed to 
        'auto' must be of the same type for the same enumeration!"""
        clsName = self.space.getClassName()
        actName = type(value).__name__
        expName = valueType.__name__
        e = es % (clsName, name, actName, expName)
        raise NotImplementedError(e)
    existing = self.space.get('__future_entries__', {})
    if name in existing:
      raise NotImplementedError('duplicate keenum: %s' % name)
    existing[name] = value
    self.space['__future_entries__'] = existing

  def _setValueType(self, valueType: type, ) -> None:
    """
    Set the value type for the KeeSpace namespace.
    """
    if '__future_value_type__' in self.space:
      existingType = self.space['__future_value_type__']
      if valueType is not existingType:
        es = """Found inconsistent value type for members of the '%s' class 
        under construction. The value type was previously set to: '%s', 
        while the new value type is: '%s'. Values passed to 'auto' must be 
        of the same type for the same enumeration!"""
        clsName = self.space.getClassName()
        e = es % (clsName, existingType.__name__, valueType.__name__)
        raise NotImplementedError(e)
      return
    self.space['__future_value_type__'] = valueType

  def _getValueType(self) -> type:
    """
    Get the value type for the KeeSpace namespace.
    """
    return self.space.get('__future_value_type__', None)

  def setItemHook(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """
    The setItemHook method is called when an item is set in the
    namespace.
    """
    if isinstance(value, AbstractObject):
      if hasattr(value, '__auto_member__'):
        if getattr(value, '__auto_member__'):
          if TYPE_CHECKING:
            assert isinstance(value, _AutoMember)
          self._addMember(key.upper(), value.getValue())
          return True
    return False

  def preCompileHook(self, compiledSpace: dict) -> dict:
    """Hook for preCompile. This is called before the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    futureEntries = compiledSpace.get('__future_entries__', None)
    if futureEntries is None:
      compiledSpace['__future_entries__'] = dict()
    return compiledSpace



"""QuestionableSyntax is raised when a name is encountered that is likely
a typo, such as '__set_item__' instead of '__setitem__' or '__setname__'
instead of '__set_name__'.
This is a"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable, Self


class QuestionableSyntax(SyntaxError):
  """QuestionableSyntax is raised when a name is encountered that is likely
  a typo, such as '__set_item__' instead of '__setitem__' or '__setname__'
  instead of '__set_name__'.
  This is a subclass of SyntaxError and should be used to indicate that
  the code is likely to be incorrect. """

  derpName = _Attribute()
  realName = _Attribute()

  def __init__(self, realName: str, derpName: str, ) -> None:
    self.__derp_name__ = derpName
    self.__real_name__ = realName
    info = """Received name: '%s' which is similar enough to '%s' to be a 
    likely typo. """
    SyntaxError.__init__(self, monoSpace(info % (derpName, realName)))

  def _resolveOther(self, other: object) -> Any:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the QuestionableSyntax object with another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    cls = type(self)
    if isinstance(other, cls):
      if self.derpName != other.derpName:
        return False
      if self.realName != other.realName:
        return False
      return True
    return False



"""EZHook collects the field entries in EZData class bodies. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType

from ..mcls.hooks import AbstractHook
from ..text import stringList
from ..waitaminute import attributeErrorFactory, ReservedName

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Iterator
  from worktoy.ezdata import EZData


class EZHook(AbstractHook):
  """EZHook collects the field entries in EZData class bodies. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __reserved_names__ = """__slots__, __init__, __eq__, __str__, __repr__,
    __iter__, __getitem__, __setitem__, __getattr__"""
  __ignore_names__ = """__module__, __dict__, __weakref__, __class__, 
  __qualname__, __firstlineno__, __doc__, __static_attributes__, 
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  @classmethod
  def _getReservedNames(cls) -> list[str]:
    """Get the reserved names for the EZData class."""
    return stringList(cls.__reserved_names__)

  @classmethod
  def _getIgnoreNames(cls) -> list[str]:
    """Get the ignore names for the EZData class."""
    return stringList(cls.__ignore_names__)

  @classmethod
  def validateName(cls, name: str, ) -> None:
    """
    Raises ReservedName if the name is one of the reserved names.
    """
    if name in cls._getReservedNames():
      raise ReservedName(name)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Hook methods

  def preCompileHook(self, compiledSpace: dict) -> dict:
    """The preCompileHook method is called before the class is compiled."""
    dataFields = self.space.getDataFields()
    compiledSpace['__slots__'] = (
        *[dataField.key for dataField in dataFields],)
    return compiledSpace

  def postCompileHook(self, compiledSpace: dict) -> dict:
    """The postCompileHook method is called after the class is compiled."""
    dataFields = self.space.getDataFields()
    initMethod = self.initFactory(*dataFields)
    eqMethod = self.eqFactory(*dataFields)
    strMethod = self.strFactory(*dataFields)
    reprMethod = self.reprFactory(*dataFields)
    iterMethod = self.iterFactory(*dataFields)
    getItemMethod = self.getItemFactory(*dataFields)
    setItemMethod = self.setItemFactory(*dataFields)
    getAttrMethod = self.getAttrFactory(*dataFields)
    compiledSpace['__init__'] = initMethod
    compiledSpace['__eq__'] = eqMethod
    compiledSpace['__str__'] = strMethod
    compiledSpace['__repr__'] = reprMethod
    compiledSpace['__iter__'] = iterMethod
    compiledSpace['__getitem__'] = getItemMethod
    compiledSpace['__setitem__'] = setItemMethod
    compiledSpace['__getattr__'] = getAttrMethod
    return compiledSpace

  def setItemHook(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """The setItemHook method is called when an item is set in the
    enumeration."""
    if key in self._getIgnoreNames():
      return False
    if key in self._getReservedNames():
      if hasattr(value, '__is_root__'):
        return False
      raise ReservedName(key)
    if callable(value):
      return False
    self.space.addField(key, type(value), value)
    return True

  # \_____________________________________________________________________/ #
  #  Method factories
  # /¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨\ #
  @staticmethod
  def initFactory(*dataFields) -> FunctionType:
    """
    Creates the '__init__' method for the EZData class.
    """

    slotKeys = [dataField.key for dataField in dataFields]
    defVals = [dataField.val for dataField in dataFields]

    def __init__(self, *args, **kwargs):
      """
      The generated '__init__' method sets attributes on the instance
      based on given arguments. Keyword arguments take precedence.
      Positional arguments are applied in order.
      """
      posArgs = [*args, ]
      while len(posArgs) < len(slotKeys):
        posArgs.append(None)
      for key, defVal in zip(slotKeys, defVals):
        setattr(self, key, defVal)
      for key, arg in zip(slotKeys, posArgs):
        if arg is not None:
          setattr(self, key, arg)
      for key in slotKeys:
        if key in kwargs:
          setattr(self, key, kwargs[key])

    return __init__

  @staticmethod
  def eqFactory(*dataFields) -> FunctionType:
    """
    Creates the '__eq__' method for the EZData class.
    """

    def __eq__(self, other: EZData) -> bool:
      """
      Instances of EZData are equal if each of their data fields are equal.
      """
      for dataField in dataFields:
        key = dataField.key
        if getattr(self, key) != getattr(other, key):
          return False
      return True

    return __eq__

  @staticmethod
  def strFactory(*dataFields) -> FunctionType:
    """The strFactory method is called when the class is created."""

    def __str__(self) -> str:
      """The __str__ method is called when the class is created."""
      clsName = type(self).__name__
      keys = [dataField.key for dataField in dataFields]
      vals = [str(getattr(self, key)) for key in keys]
      keyVals = ['%s=%s' % (key, val) for key, val in zip(keys, vals)]
      return """%s(%s)""" % (clsName, ', '.join(keyVals))

    return __str__

  @staticmethod
  def reprFactory(*dataFields) -> FunctionType:
    """The reprFactory method is called when the class is created."""

    def __repr__(self) -> str:
      """The __repr__ method is called when the class is created."""
      clsName = type(self).__name__
      keys = [dataField.key for dataField in dataFields]
      vals = [str(getattr(self, key)) for key in keys]
      return """%s(%s)""" % (clsName, ', '.join(vals))

    return __repr__

  @staticmethod
  def iterFactory(*dataFields) -> FunctionType:
    """The iterFactory method is called when the class is created."""

    def __iter__(self, ) -> Iterator:
      """
      Implementation of the iteration protocol
      """
      for key in self.__slots__:
        yield getattr(self, key)

    return __iter__

  @staticmethod
  def getItemFactory(*dataFields) -> FunctionType:
    """The getItemFactory method is called when the class is created."""

    def __getitem__(self, key: str) -> object:
      """The __getitem__ method is called when the class is created."""
      if key in self.__slots__:
        return getattr(self, key)
      raise KeyError(key)

    return __getitem__

  @staticmethod
  def setItemFactory(*dataFields) -> FunctionType:
    """The setItemFactory method is called when the class is created."""

    def __setitem__(self, key: str, value: object) -> None:
      """The __setitem__ method is called when the class is created."""
      if key in self.__slots__:
        return setattr(self, key, value)
      raise KeyError(key)

    return __setitem__

  @staticmethod
  def getAttrFactory(*dataFields) -> FunctionType:
    """The getAttrFactory method is called when the class is created."""

    def __getattr__(self, key: str) -> Any:
      """The __getattr__ method is called when the class is created."""
      if key in self.__slots__:
        return getattr(self, key)
      raise attributeErrorFactory(type(self), key)

    return __getattr__



"""The 'worktoy.attr' module implements the descriptor protocol."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._field import Field
from ._attri_box import AttriBox

__all__ = [
    'Field',
    'AttriBox',
]



"""
HistDict is a subclass of dict that keeps track of all accessor calls made
to either of its '__getitem__', '__setitem__', and '__delitem__' methods.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..parse import maybe
from . import ItemCall
from ..waitaminute import TypeException

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class HistDict(dict):
  """
  HistDict is a subclass of dict that keeps track of all accessor calls made
  to either of its '__getitem__', '__setitem__', and '__delitem__' methods.
  """
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __item_calls__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getItemCalls(self) -> list[ItemCall]:
    """
    Returns a list of all item calls made to the namespace.
    """
    return maybe(self.__item_calls__, [])

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def addItemCall(self, *args) -> None:
    """
    Adds an item call to the namespace.
    """
    if len(args) == 1 and isinstance(args[0], ItemCall):
      itemCall = args[0]
    else:
      itemCall = ItemCall(*args)
    existing = self.getItemCalls()
    self.__item_calls__ = [*existing, itemCall]

  def extendItemCalls(self, *args, **kwargs) -> None:
    """
    Extends the item calls with the provided list of ItemCall objects.
    """
    cls = type(self)
    itemCalls = []
    other = None
    for arg in args:
      if isinstance(arg, cls):
        other = arg
        break
      elif isinstance(arg, ItemCall):
        itemCalls.append(arg)
    else:
      for itemCall in itemCalls:
        if isinstance(itemCall, ItemCall):
          self.addItemCall(itemCall)
      else:
        return None
    if kwargs.get('_recursion', False):
      raise RecursionError
    return self.extendItemCalls(*other.getItemCalls(), _recursion=True)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __getitem__(self, key: str, **kwargs) -> Any:
    """
    Gets the value associated with the key in the namespace.
    """
    try:
      value = dict.__getitem__(self, key)
    except KeyError as keyError:
      self.addItemCall(key, keyError, dict.__getitem__)
      raise keyError
    else:
      self.addItemCall(key, value, dict.__getitem__)
      return value
    finally:
      pass

  def __setitem__(self, key: str, value: Any, **kwargs) -> None:
    """
    Sets the value associated with the key in the namespace.
    """
    self.addItemCall(key, value, dict.__setitem__)
    return dict.__setitem__(self, key, value)

  def __delitem__(self, key: str, **kwargs) -> None:
    """
    Deletes the value associated with the key in the namespace.
    """
    try:
      value = dict.__getitem__(self, key)
    except KeyError as keyError:
      self.addItemCall(key, keyError, dict.__delitem__)
    else:
      self.addItemCall(key, value, dict.__delitem__)
      return dict.__delitem__(self, key)
    finally:
      pass

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args: Any, **kwargs: Any) -> None:
    """
    Initializes the HistDict instance and sets up the item calls list.
    """
    dict.__init__(self, )
    for key, val in kwargs.items():
      if isinstance(key, str):
        raise TypeException('key', key, str)
      self[key] = val



"""FlexMismatch is a custom exception indicating failure to flexible cast
positional arguments to a TypeSig object."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import CastMismatch


class FlexMismatch(CastMismatch):
  """
  FlexMismatch is a custom exception indicating failure to flexible cast
  positional arguments to a TypeSig object.
  """
  pass



"""TypeException is a custom exception class for handling type related
errors. Specifically, this exception should NOT be raised if the object is
None instead of the expected type. This is because None indicates absense
rather than type mismatch. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace, joinWords

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


def _resolveTypeNames(*types) -> str:
  """Creates the first part of the error message listing the expected type
  or types. """
  if len(types) == 1:
    if isinstance(types[0], (tuple, list)):
      return _resolveTypeNames(*types[0])
    if isinstance(types[0], type):
      expName = types[0].__name__
    elif isinstance(types[0], str):
      expName = types[0]
    else:
      raise TypeError("""Received bad arguments: %s""" % (str(types),))
    return """Expected object of type '%s'""" % (expName,)
  typeNames = []
  for type_ in types:
    if isinstance(type_, type):
      typeNames.append("""'%s'""" % type_.__name__)
    elif isinstance(type_, str):
      typeNames.append("""'%s'""" % type_)
    else:
      raise TypeError("""Received bad arguments: %s""" % (str(types),))
  infoSpec = """Expected object of any of the following types: %s"""
  typeStr = joinWords(*typeNames, sep='or')
  return monoSpace(infoSpec % (typeStr,))


class _Meta(type):
  """Metaclass simplifying __str__ and __repr__"""

  def __str__(cls) -> str:
    """String representation of the class."""
    return cls.__name__

  def __repr__(cls) -> str:
    """Representation of the class."""
    return cls.__name__


class TypeException(TypeError):
  """
  TypeException is a custom exception class for handling type related
  errors. Specifically, this exception should NOT be raised if the object is
  None instead of the expected type.
  """

  varName = _Attribute()
  actualObject = _Attribute()  # This is the object that was received
  actualType = _Attribute()
  expectedType = _Attribute()

  def __init__(self, name: str, obj: object, *types) -> None:
    """Initialize the TypeException with the name of the variable, the
    received object, and the expected types."""
    TypeError.__init__(self, )
    self.varName = name
    self.actualObject = obj
    self.actualType = type(obj)
    self.expectedType = types

  def __str__(self) -> str:
    """String representation of the TypeException."""
    prelude = _resolveTypeNames(*self.expectedType)
    actName = type(self.actualObject).__name__
    actRepr = repr(self.actualObject)
    infoSpec = """%s at name: '%s', but received object of type '%s' with 
    repr: '%s'"""
    info = infoSpec % (prelude, self.varName, actName, actRepr)
    return monoSpace(info)

  __repr__ = __str__



"""AutoCast subclasses AbstractCast and provides a general casting utility
for arbitrary types. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import AbstractCast

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class AutoCast(AbstractCast):
  """AutoCast subclasses AbstractCast and provides a general casting
  utility for arbitrary types. """

  __target_type__ = None

  def getTargetType(self, **kwargs) -> type:
    """Get the target type of the cast. Subclasses must implement this
    method to return the target type."""
    return self.__target_type__

  def _setTargetType(self, targetType: type) -> None:
    """Set the target type of the cast. This method is used to set the
    target type of the cast."""
    self.__target_type__ = targetType

  def _cast(self, *args, **kwargs) -> Any:
    """Cast the arguments to the target type."""
    if kwargs.get('_recursion', False):
      raise RecursionError
    cls = self.getTargetType()
    return cls(*args, **kwargs, _recursion=True)

  def __init__(self, targetType: type) -> None:
    """Initialize the AutoCast object."""
    self._setTargetType(targetType)



"""
'OWNER' provides a companion to 'THIS' allowing references to both a class
and an instance of a class respectively. This relationship is emphasized,
is reflected by 'THIS' being recognized by 'OWNER' as an instance of it.

For example:

import sys

def main() -> int:
  if isinstance(THIS, OWNER):
    return 0
  return 1

if __name__ == '__main__':
  sys.exit(main())  # Returns 0, as 'THIS' is an instance of 'OWNER'
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...text import monoSpace

from . import Zeroton, THIS


class OWNER(metaclass=Zeroton):
  """Similar to THIS, but referring to the bound class object instead. """

  @classmethod
  def __class_str__(cls, ) -> str:
    """Return the class name of the class."""
    infoSpec = """The '%s' Zeroton is a placeholder for the class object 
    of the surrounding scope. When used in the same context as 'THIS', 
    it refers to the class object of the surrounding scope and 'THIS' 
    refers to the instance of the class."""
    return monoSpace(infoSpec % cls.__name__)

  @classmethod
  def __class_instancecheck__(cls, obj: object) -> bool:
    """When 'OWNER' and 'THIS' occurs in the same context, generally,
    'THIS' is a placeholder for an instance of the class for which 'OWNER'
    is a placeholder. """
    return True if cls is obj or obj is THIS else False



"""
BaseObject is the standard entry point for using the worktoy library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import BaseMeta
from ..static import AbstractObject

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False


class BaseObject(AbstractObject, metaclass=BaseMeta):
  """
  BaseObject combines the core functionality of AbstractObject with the
  hook-based metaclass behavior of BaseMeta.

  From AbstractObject, it inherits robust constructor handling, controlled
  descriptor mutation, and context-aware access to instance and owner
  information. From BaseMeta, it gains support for function overloading
  and other hook-based class construction features via BaseSpace.

  Subclass this when you want both: sane, safe object semantics and
  overload-aware metaclass support.
  """
  pass



"""The stringList function receives a single string describing a list of
items and returns a list of strings each representing an item."""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from . import monoSpace


def stringList(items: str, **kwargs) -> list[str]:
  """The stringList function receives a single string describing a list of
  items and returns a list of strings each representing an item."""
  if not items:
    return []
  separator = kwargs.get('separator', ',')
  items = monoSpace(items).split(separator)
  items = [item.strip() for item in items]
  return items



"""Cast provides a flexible casting utility used by the overload system. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import Any


class AbstractCast:
  """Cast provides a flexible casting utility used by the overload
  system. """

  @abstractmethod
  def getTargetType(self, **kwargs) -> type:
    """Get the target type of the cast. Subclasses must implement this
    method to return the target type."""

  @abstractmethod
  def _cast(self, *args, **kwargs) -> Any:
    """Cast the arguments to the target type."""

  def __call__(self, *args, **kwargs) -> Any:
    """Calling the Cast object will cast the arguments to the target type."""
    if kwargs.get('_recursion', False):
      raise RecursionError
    return self._cast(*args, **kwargs, _recursion=True)



"""The 'worktoy.static.casting' module provides type casting."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._abstract_cast import AbstractCast
from ._auto_cast import AutoCast
from ._int_cast import IntCast
from ._float_cast import FloatCast
from ._complex_cast import ComplexCast
from ._cast import Cast

__all__ = [
    'AbstractCast',
    'AutoCast',
    'IntCast',
    'FloatCast',
    'ComplexCast',
    'Cast',
]



"""The 'validateExistingFile' function validates the existence of a file. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Union, Self, Callable, TypeAlias


def validateExistingFile(file: str) -> str:
  """
  Validates that a given 'str' object points to an existing file.

  Args:
    file (str): The file to validate.

  Returns:
    str: The validated file.

  Raises:
    FileNotFoundError: If the file does not exist.
    IsADirectoryError: If the path is a directory.
  """
  if not os.path.exists(file):
    infoSpec = """No file exists at: '%s'!"""
    info = monoSpace(infoSpec % file)
    raise FileNotFoundError(info)
  if not os.path.isfile(file):
    infoSpec = """The path '%s' is not a file!"""
    info = monoSpace(infoSpec % file)
    raise IsADirectoryError(info)
  return os.path.normpath(file)



"""MissingVariable exception should be raised when a variable is missing
and requires initialization. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import joinWords

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self


class MissingVariable(Exception):
  """MissingVariable exception should be raised when a variable is missing
  and requires initialization. """

  varName = _Attribute()
  varType = _Attribute()

  def __init__(self, name: str, *types) -> None:
    """Initialize the MissingVariable object."""
    self.varName = name
    self.varType = types
    if not types:
      info = """Missing variable at name: '%s'!%s"""
      typeStr = ''
    elif len(types) == 1:
      info = """Missing variable '%s' of type '%s'"""
      typeStr = types[0].__name__
    else:
      info = """Missing variable '%s' of any of the following types: %s"""
      typeNames = [type_.__name__ for type_ in types]
      typeStr = joinWords(*typeNames, sep='or')
    Exception.__init__(self, info % (name, typeStr))

  def _resolveOther(self, other: object) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the MissingVariable object with another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    cls = type(self)
    if isinstance(other, cls):
      if self.varName != other.varName:
        return False
      if len(self.varType) != len(other.varType):
        return False
      for selfType, otherType in zip(self.varType, other.varType):
        if selfType != otherType:
          return False
      return True
    return False



"""
KeeMeta provides the metaclass creating the KeeNum enumeration class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..attr import Field
from ..mcls import AbstractMetaclass
from ..parse import maybe
from ..static import AbstractObject
from ..waitaminute import MissingVariable, TypeException
from ..waitaminute import IllegalInstantiation
from . import KeeSpace, _KeeNumBase

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, TypeAlias, Any, Iterator

  Bases: TypeAlias = tuple[type, ...]
  MTypes: TypeAlias = dict[str, type]

try:
  from icecream import ic  # NOQA

  ic.configureOutput(includeContext=True, )
except ImportError:
  print('icecream not installed, using dummy ic function.')
  ic = print


class KeeMeta(AbstractMetaclass):
  """
  KeeMeta provides the metaclass creating the KeeNum enumeration class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __core_keenum__ = None  # Future KeeNum base class

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _createCoreKeeNum(mcls, ) -> Self:
    """
    Creator function for the core KeeNum class.
    """

    class KeeNum(_KeeNumBase, metaclass=mcls):
      """
      KeeNum provides the base class for enumerating classes with
      restricted and predefined instances called members.
      """
      pass

    setattr(mcls, '__core_keenum__', KeeNum)

  @classmethod
  def getCoreKeeNum(mcls, **kwargs) -> type:
    """
    Get the core KeeNum class.
    """
    if mcls.__core_keenum__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      mcls._createCoreKeeNum()
      return mcls.getCoreKeeNum(_recursion=True)
    if isinstance(mcls.__core_keenum__, type):
      return mcls.__core_keenum__
    name, value = '__core_keenum__', mcls.__core_keenum__
    raise TypeException(name, value, type)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> KeeSpace:
    """
    Prepare the class namespace.
    """
    return KeeSpace(mcls, name, bases, **kwargs)

  def __new__(mcls, name: str, bases: Bases, space: KeeSpace, **kw) -> Self:
    """
    Create a new instance of the KeeMeta metaclass.
    """
    if mcls.__core_keenum__ is None:
      if name == 'KeeNum' or kw.get('_root', False):
        return super().__new__(mcls, name, bases, space, **kw)
      raise RuntimeError
    coreKeeNum = mcls.getCoreKeeNum()
    return super().__new__(mcls, name, (coreKeeNum,), space, **kw)

  def __init__(cls, name: str, bases: Bases, space: KeeSpace, **kw) -> None:
    """
    Initialize the KeeMeta metaclass.
    """
    super().__init__(name, bases, space, **kw)
    if name == 'KeeNum':
      return
    futureEntries = getattr(cls, '__future_entries__', )
    if futureEntries is None:
      raise MissingVariable('__future_entries__', dict)
    if not isinstance(futureEntries, dict):
      raise TypeException('__future_entries__', futureEntries, dict)
    actualEntries = dict()
    valueType = getattr(cls, '__future_value_type__', )
    memberValues = []
    memberNames = []
    for i, (name, value) in enumerate(futureEntries.items()):
      entry = cls(_root=True)
      entry.__member_name__ = name
      entry.__member_value__ = value
      entry.__member_index__ = i
      memberValues.append(value)
      memberNames.append(name)
      actualEntries[name] = entry
      setattr(cls, name, entry)
      if name != name.upper():
        setattr(cls, name.upper(), entry)

    cls._validate(**actualEntries)

    setattr(cls, '__value_type__', valueType)
    setattr(cls, '__member_entries__', actualEntries)
    delattr(cls, '__future_entries__')
    delattr(cls, '__future_value_type__')

  def __call__(cls, *args, **kwargs) -> Any:
    """
    Only the _addMember method is allowed to create instances of the
    KeeNum class.
    """
    if not kwargs.get('_root', False):
      raise IllegalInstantiation(cls)
    self = super().__call__(**kwargs)
    return self

  def __len__(cls) -> int:
    """
    Get the number of members in the KeeNum class.
    """
    out = 0
    for _ in cls:
      out += 1
    else:
      return out

  def __iter__(cls, ) -> Iterator:
    """
    Iterate over the members of the KeeNum class.
    """
    yield from getattr(cls, '__member_entries__', ).values()

  def __getitem__(cls, key: str) -> Any:
    """
    Get a member by its name.
    """
    valueType = getattr(cls, '__value_type__', None)
    if valueType is int and isinstance(key, int):
      try:
        out = cls._getFromValue(key)
      except ValueError as valueError:
        try:
          out = cls._getFromIndex(key)
        except IndexError as indexError:
          raise indexError from valueError
        else:
          return out
      else:
        return out
    if valueType is str and isinstance(key, str):
      try:
        out = cls._getFromValue(key)
      except ValueError as valueError:
        try:
          out = cls._getFromName(key)
        except KeyError as keyError:
          raise keyError from valueError
        else:
          return out
      else:
        return out
    if isinstance(key, int):
      return cls._getFromIndex(key)
    if isinstance(key, str):
      return cls._getFromName(key)
    try:
      out = cls._getFromValue(key)
    except ValueError as valueError:
      raise KeyError(key) from valueError
    else:
      return out

  def _getFromIndex(cls, index: int) -> Any:
    """
    Get a member by its index. If the value type is 'int' and the index is
    negative, this method assumes that the index does not match any value
    type.
    """
    if index < 0:
      return cls._getFromIndex(len(cls) + index)
    if index >= len(cls):
      raise IndexError(f'Index {index} out of range for {cls.__name__}')
    for entry in cls:
      if entry.index == index:
        return entry
    raise IndexError(f'No member with index {index} in {cls.__name__}')

  def _getFromName(cls, name: str, **kwargs) -> Any:
    """
    Get a member by its name. If the value type is 'str', this method
    assumes that the identifier does not match any value type.
    """
    if not isinstance(name, str):
      raise TypeException('name', name, str)
    for entry in cls:
      if entry.name == name:
        return entry
    targetNames = [
        name,
        name.lower(),
        name.upper(),
    ]
    for entry in cls:
      entryNames = [
          entry.name,
          entry.name.lower(),
          entry.name.upper(),
      ]
      for i, j in zip(targetNames, entryNames):
        if i == j:
          return entry
    raise KeyError(name)

  def _getFromValue(cls, value: Any, **kwargs) -> Any:
    """
    Get a member by its value.
    """
    for entry in cls:
      if entry.value == value:
        return entry
    raise ValueError(value)

  def __str__(cls) -> str:
    """
    String representation of the KeeMeta metaclass.
    """
    if cls.__name__ == 'KeeNum':
      return object.__str__(cls)
    valueType = getattr(cls, '__value_type__', )
    typeName = '' if valueType is str else valueType.__name__
    clsName = cls.__name__
    if valueType is str:
      infoSpec = """%s%s(KeeNum)"""
    else:
      infoSpec = """%s(KeeNum)[%s]"""
    return infoSpec % (clsName, typeName)

  def __repr__(cls) -> str:
    """
    Representation of the KeeMeta metaclass.
    """
    return object.__repr__(cls)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @staticmethod
  def _validateStrValues(**entries) -> None:
    """
    Validates that exactly one of the two following conditions is satisfied:
    1. Every name is the same as the member value
    2. No name is the name of a member value
    """
    memberValueNames = []
    for name, value in entries.items():
      if not isinstance(name, str):
        raise TypeException('name', name, str)
      memberValueNames.append(value)
      memberValueNames.append(value.lower())
      memberValueNames.append(value.upper())
    for name, value in entries.items():
      if name not in memberValueNames:
        break
    else:  # Means 'break' was never hit
      return
    for name, value in entries.items():
      if name in memberValueNames:
        break
    else:  # Means 'break' was never hit
      return
    raise NotImplementedError('name inconsistency: %s' % name)

  @staticmethod
  def _validateIntValues(**entries) -> None:
    """
    Validates that not value would match the index of a member or be
    negative.
    """
    for name, value in entries.items():
      if not isinstance(value, int):
        raise TypeException('value', value, int)
      if value < 0:
        raise ValueError('Negative values are not allowed: %s' % value)
      if value < len(entries):
        break
    else:  # Means 'break' was never hit
      return
    raise NotImplementedError('value inconsistency: %s' % value)

  @staticmethod
  def _validateValues(**entries) -> None:
    """
    Validates type consistency of the values in the entries. Please note
    that this method does not apply to validation of types 'int' and
    'str'. For those cases, use the methods '_validateIntValues' and
    '_validateStrValues' respectively.
    """
    valueType = None
    for name, value in entries.items():
      if valueType is None:
        valueType = type(value)
        continue
      if not isinstance(value, valueType):
        break
    else:  # Means 'break' was never hit
      return
    raise TypeException('value', value, valueType)

  @staticmethod
  def _validate(**entries) -> None:
    """
    Dispatches entries to the appropriate validation method based on
    the type of the values.
    """
    if not entries:
      return
    valueType = None
    for name, value in entries.items():
      valueType = type(value)
      break
    if valueType is str:
      KeeMeta._validateStrValues(**entries)



"""
The ClassBodyTemplate provides a string template for the body of a class
body.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, TypeAlias, Any

ClassBodyTemplate = """
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  
  #  Fallback Variables
  
  #  Private Variables
  
  #  Public Variables
  
  #  Virtual Variables

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  
    Method decorators in class bodies are certain to encounter function
    objects rather than bound methods, but this applies only to the
    immediately applied decorator. Decorators applied to already decorated
    methods receive instead the object return by the previous decorator.

    A typical decorator pattern is for the decorator to wrap the decorated
    function object at the '__wrapped__' attribute. Next, the decorator
    class should then implement '__call__' to forward the call to the
    wrapped method.

    An improvement of the above pattern is to emulate the bound method
    behaviour by implementing the descriptor protocol on the decorator
    class. This requires implementing '__get__' to return the wrapped
    function body directly, when the instance argument is 'None' and to
    return a wrapping function that already has the instance at the first
    argument position when the instance argument is not 'None'.

    The most elegant pattern however, is for the decorator to be
    instantiated on the class body separately from the decorated method.
    When decorating the method, the decorator should only record a
    reference to the method or even just the name of the method, before
    returning the decorator instance having applied no changes to it.
    Decorators following this pattern stacks without having to implement
    inelegant workarounds to handle the cases where they decorate another
    decorator.

    IDEA:
    Implement a namespace hooks that recognizes when a decorator is
    getting called and then to allow the hook to handle the control flow
    of applying the decorator!
"""



"""
The Dispatch class dispatches a function call to the appropriate
function based on the type of the first argument.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os
from types import FunctionType as Func
from types import MethodType as Meth

from . import AbstractObject
from ..attr import Field
from ..static import TypeSig
from ..text import typeMsg, monoSpace
from ..waitaminute import HashMismatch, CastMismatch, FlexMismatch
from ..waitaminute import DispatchException, CascadeException
from ..waitaminute import IllegalDispatcher, MissingVariable

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable, TypeAlias

  Types: TypeAlias = tuple[type, ...]
  Hashes: TypeAlias = list[int]
  HashMap: TypeAlias = dict[int, Callable]
  TypesMap: TypeAlias = dict[Types, Callable]
  CastMap: TypeAlias = dict[Types, Callable]
  CallMap: TypeAlias = dict[TypeSig, Callable]


class Dispatch(AbstractObject):
  """
Dispatch replaces the usual bound method when overloaded function
objects are used. The Dispatch instance serves as a dynamic method
selector, attached to a class attribute in place of the original
function object.

Dispatch achieves this by subclassing AbstractObject, which provides
a full implementation of the descriptor protocol, including
__get__, __set__, and __delete__. This allows Dispatch to operate as
a descriptor that controls method binding and function dispatching.

When called, Dispatch attempts to match the provided arguments to one
of its registered TypeSig signatures. It proceeds in the following stages:

1. Fast dispatch:
   The most performant stage. Arguments must match the expected types
   exactly — even an int will not match a float. If an exact match is
   found, Dispatch immediately invokes the corresponding function.

2. Cast dispatch:
   If fast dispatch fails, Dispatch attempts to cast arguments to the
   required types, proceeding only if all casts succeed.

3. Flex dispatch:
   If casting also fails, Dispatch performs flexible matching. It may
   reorder arguments and unpack iterables as needed to satisfy the
   signature.

Classes derived from BaseMeta can decorate methods with @overload([TYPES])
to indicate that the decorated object should be dispatched when receiving
arguments matching the given types. The custom metaclass control flow then
instantiates Dispatch during class creation.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class variables
  __latest_dispatch__ = None  # The latest dispatch that was made
  __running_tests__ = None  # Whether the class is running tests
  __overload_dispatcher__ = True  # Required flag for all dispatchers!

  #  Private variables
  __call_map__ = None

  #  Public variables
  __name__ = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getTypeSigs(self) -> list[TypeSig]:
    """
    Getter-function for the type signatures supported.
    """
    return [*self.__call_map__.keys(), ]

  @classmethod
  def getLatestDispatch(cls) -> Func:
    """
    Getter-function for the most recently successful dispatch.
    """
    if cls.__latest_dispatch__ is None:
      raise MissingVariable('__latest_dispatch__', Meth)
    return cls.__latest_dispatch__.__func__

  @classmethod
  def _createTestFlag(cls) -> None:
    """
    Create the test flag for the class.
    """
    value = os.environ.get('RUNNING_TESTS', '')
    cls.__running_tests__ = True if value else False

  @classmethod
  def getTestFlag(cls, **kwargs) -> bool:
    """
    Get the test flag.
    """
    if cls.__running_tests__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      cls._createTestFlag()
      return cls.getTestFlag(_recursion=True)
    return cls.__running_tests__

  @__name__.GET
  def _getName(self, ) -> str:
    """
    Get the name of the function.
    """
    return self.getFieldName()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _resetLatestDispatch(cls) -> None:
    """
    Reset the latest dispatch to None.
    """
    cls.__latest_dispatch__ = None

  @classmethod
  def _setLatestDispatch(cls, dispatch: Callable) -> None:
    """
    Set the latest dispatch to the given dispatch.
    """
    cls.__latest_dispatch__ = dispatch

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, callMap: CallMap) -> None:
    self.__call_map__ = callMap

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: type, name: str, **kwargs) -> None:
    """
    When the owning class is created, Python calls this method to allowing
    the type signatures to be updated with the owner class. This is
    necessary as the type signatures are able to reference the owning
    class before it is created by using the 'THIS' token object in place
    of it.
    """
    AbstractObject.__set_name__(self, owner, name, **kwargs)

    for sig, call in self.__call_map__.items():
      if not isinstance(sig, TypeSig):
        raise TypeError(typeMsg('sig', sig, TypeSig))
      if not callable(call):
        raise TypeError(typeMsg('call', call, Func))
      TypeSig.replaceTHIS(sig, owner)

  def __init_subclass__(cls, **kwargs) -> None:
    """
    This method checks that subclasses retain the token flag to indicate
    their use as dispatchers. See the 'IllegalDispatcher' documentation
    for more details.
    """
    try:
      _ = cls.__overload_dispatcher__
    except Exception as exception:
      raise IllegalDispatcher(cls, ) from exception

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _fastDispatch(self, *args, **kwargs) -> Any:
    """
    Fast dispatch the function call.
    """
    exceptions = []
    for sig, call in self.__call_map__.items():
      try:
        posArgs = sig.fast(*args)
      except HashMismatch as hashMismatch:
        exceptions.append(hashMismatch)
        continue
      else:
        if hasattr(call, '__func__'):
          return call.__func__(self.instance, *posArgs, **kwargs)
        return call(self.instance, *posArgs, **kwargs)
    else:
      if exceptions:
        raise exceptions[-1]  # Will be caught by control flow
      raise RuntimeError("""No signatures defined!""")

  def _castDispatch(self, *args, **kwargs) -> Any:
    """
    Dispatches the function call with arguments casted to the expected
    types.
    """
    exceptions = []
    for sig, call in self.__call_map__.items():
      try:
        posArgs = sig.cast(*args)
      except CastMismatch as castMismatch:
        exceptions.append(castMismatch)
        continue
      else:
        return call(self.instance, *posArgs, **kwargs)
    else:
      if exceptions:
        raise exceptions[-1]  # Will be caught by control flow
      raise RuntimeError("""No signatures defined!""")

  def _flexDispatch(self, *args, **kwargs) -> Any:
    """
    The most flexible attempt to dispatch the function call.
    """
    exceptions = []
    for sig, call in self.__call_map__.items():
      try:
        posArgs = sig.flex(*args)
      except Exception as exception:
        exceptions.append(exception)
        continue
      else:
        return call(self.instance, *posArgs, **kwargs)
    else:
      if exceptions:
        raise exceptions[-1]  # Will be caught by control flow
      raise RuntimeError("""No signatures defined!""")

  def _dispatch(self, *args: Any, **kwargs: Any) -> Any:
    """
    Dispatches the function call by trying fast, cast and flex in that
    order.
    """
    testFlag = self.getTestFlag()
    if testFlag:
      self._setLatestDispatch(self._fastDispatch)
    exceptions = []
    try:
      out = self._fastDispatch(*args, **kwargs)
    except HashMismatch as hashMismatch:
      exceptions.append(hashMismatch)
    else:
      return out
    try:
      out = self._castDispatch(*args, **kwargs)
    except CastMismatch as castMismatch:
      exceptions.append(castMismatch)
    else:
      if testFlag:
        self._setLatestDispatch(self._castDispatch)
      return out
    try:
      out = self._flexDispatch(*args, **kwargs)
    except FlexMismatch as flexMismatch:
      exceptions.append(flexMismatch)
    else:
      if testFlag:
        self._setLatestDispatch(self._flexDispatch)
      return out
    if testFlag:
      self._resetLatestDispatch()
    try:
      raise CascadeException(*exceptions)
    except CascadeException as cascadeException:
      raise DispatchException(self, *args) from cascadeException

  def __call__(self, *args: Any, **kwargs: Any) -> Any:
    """
    Tries fast, cast and flex dispatches in that order before raising
    'DispatchException'.
    """
    if self.getTestFlag():
      self._resetLatestDispatch()
    return self._dispatch(*args, **kwargs)

  def __str__(self, ) -> str:
    """Get the string representation of the function."""
    sigStr = [str(sig) for sig in self.getTypeSigs()]
    info = """%s object supporting type signatures: \n%s"""
    sigLines = '<br><tab>'.join(sigStr)
    return monoSpace(info % (self.__field_name__, sigLines))

  def __repr__(self, ) -> str:
    """Get the string representation of the function."""
    return object.__repr__(self)



"""
HashError provides a custom exception raised to indicate an unexpected
hash value.

Provide expected value and then actual value to the constructor. If a
single value is provided, it is understood to be the actual unexpected
value.

"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class HashError(RuntimeError):
  """
  HashError provides a custom exception raised to indicate an unexpected
  hash value.
  """

  expectedValue = _Attribute()
  actualValue = _Attribute()

  @staticmethod
  def _getInfoNoValues() -> str:
    """Return the error message without values."""
    infoSpec = """Received unexpected hash value!"""
    return infoSpec

  @staticmethod
  def _getInfoOneValue(value: int) -> str:
    """Return the error message with one value. The given value is
    understood to be the actual value, not the expected value."""
    infoSpec = """Received unexpected hash value: %d!"""
    return infoSpec % value

  @staticmethod
  def _getInfoTwoValues(expected: int, actual: int) -> str:
    """Return the error message with two values."""
    infoSpec = """Received unexpected hash value: %d! Expected: %d!"""
    return infoSpec % (actual, expected)

  def _getInfo(self, ) -> str:
    """
    Parses the values to info specification.
    """
    if self.expectedValue is None and self.actualValue is None:
      return self._getInfoNoValues()
    if self.actualValue is None:
      return self._getInfoOneValue(self.expectedValue)
    if self.expectedValue is None:
      return self._getInfoNoValues()
    return self._getInfoTwoValues(self.expectedValue, self.actualValue)

  def __init__(self, *values) -> None:
    """
    Initialize the HashError exception.

    Args:
      values: int
        The expected and actual hash values.
    """
    intArgs = [i for i in values if isinstance(i, int)]
    self.expectedValue, self.actualValue = [*intArgs, None, None][:2]
    RuntimeError.__init__(self, self._getInfo())

  def __eq__(self, other: Any) -> bool:
    """
    Check if the other object is a HashError and has the same values.
    """
    cls = type(self)
    if not isinstance(other, cls):
      return False
    if self.expectedValue is None and self.actualValue is None:
      return True if self is other else False
    if self.actualValue is None and other.actualValue is None:
      return True if self.expectedValue == other.expectedValue else False
    if self.expectedValue is None and other.expectedValue is None:
      return True if self.actualValue == other.actualValue else False
    return False



"""
AbstractHook provides an abstract baseclass for hooks used by the
namespaces in the metaclass system.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...static import AbstractObject, Alias

try:
  from typing import TYPE_CHECKING, Type
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable
  from worktoy.mcls import AbstractNamespace as ASpace

  AccessorHook = Callable[[ASpace, str, Any], Any]
  CompileHook = Callable[[ASpace, dict], dict]


class AbstractHook(AbstractObject):
  """
  AbstractHook is the abstract base class for defining hook objects
  used in conjunction with AbstractNamespace. These hooks enable modular,
  stage-specific interception during class body evaluation and namespace
  compilation within the metaclass system.

  ## Purpose

  Hooks allow custom behavior to be injected into the class construction
  pipeline without modifying the namespace or metaclass core logic. They
  are used to observe and/or alter how names are accessed, assigned, or
  compiled into the final class definition.

  ## Integration

  To activate a hook, simply instantiate a subclass of AbstractHook inside
  the body of a namespace class (i.e., a subclass of AbstractNamespace).
  The descriptor protocol (`__set_name__`) ensures the hook registers
  itself with the namespace automatically at definition time.

  Example:

      class MyNamespace(AbstractNamespace):
        overloadHook = OverloadHook()
        validationHook = ReservedNameHook()

  ## Lifecycle Hook Methods

  Subclasses may override any of the following methods to participate in
  different stages of the namespace lifecycle. All are optional.

  - `setItemHook(self, key, value, oldValue) -> bool`
    Called just before a name is set in the namespace.
    Returning True blocks the default behavior.

  - `getItemHook(self, key, value) -> bool`
    Called just before a name is retrieved from the namespace.
    Returning True blocks the default behavior.

  - `preCompileHook(self, compiled: dict) -> dict`
    Called after the class body finishes executing, but before the
    namespace is finalized. May transform or replace namespace contents.

  - `postCompileHook(self, compiled: dict) -> dict`
    Called immediately before the finalized namespace is handed off to the
    metaclass. Can be used for final transformations or validation.

  ## Descriptor Behavior

  AbstractHook implements the descriptor protocol. When accessed via a
  namespace class, it is bound with the following attributes:

  - `self.space` refers to the active namespace instance.
  - `self.spaceClass` refers to the namespace class itself.

  These attributes can be used to introspect the environment the hook is
  participating in.

  ## Extension Notes

  Subclasses are expected to override only the relevant hook methods.
  If none are overridden, the hook has no effect.

  The `addHook()` method of the namespace class is automatically invoked
  during registration. Hook authors do not need to call it manually.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Public variables
  space = Alias('instance')
  spaceClass = Alias('owner')

  #  TYPE_CHECKING
  if TYPE_CHECKING:
    from . import AbstractHook
    from .. import AbstractNamespace
    assert isinstance(AbstractHook.space, AbstractNamespace)
    assert issubclass(AbstractHook.spaceClass, AbstractNamespace)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  def getItemHook(self, key: str, value: Any, ) -> bool:
    """Hook for getItem. This is called before the __getitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """

  def setItemHook(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """

  def preCompileHook(self, compiledSpace: dict) -> dict:
    """Hook for preCompile. This is called before the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    return compiledSpace

  def postCompileHook(self, compiledSpace: dict) -> dict:
    """Hook for postCompile. This is called after the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    return compiledSpace

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: type, name: str, **kwargs) -> None:
    """
    After the super call, adds one self to the namespace class as a hook
    class.
    """
    super().__set_name__(owner, name, **kwargs)
    self.spaceClass.addHook(self)



"""PathSyntaxException provides a custom exception raised to indicate that
a 'str' object is not a valid absolute path. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import monoSpace
from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Union, Self, TypeAlias, LiteralString

  Path: TypeAlias = Union[str, bytes, LiteralString]


class PathSyntaxException(ValueError):
  """
  PathSyntaxException provides a custom exception raised to indicate that
  a 'str' object is not a valid absolute path.
  """

  badPath = _Attribute()

  def __init__(self, path: Path) -> None:
    """
    Initialize the PathSyntaxException with the invalid path.

    Args:
      path (str): The invalid path.
    """
    self.badPath = path
    infoSpec = """The path '%s' is not a valid, absolute path!"""
    info = monoSpace(infoSpec % path)
    ValueError.__init__(self, info)

  def _resolveOther(self, other: Any) -> Self:
    """
    Resolve the other object to a PathSyntaxException.

    Args:
      other (Any): The other object to resolve.

    Returns:
      Self: The resolved PathSyntaxException.
    """
    cls = type(self)
    if isinstance(other, cls):
      return other
    try:
      out = cls(other)
    except (TypeError, IndexError, ValueError):
      return NotImplemented
    else:
      return out
    finally:
      if TYPE_CHECKING:  # pycharm, please!
        pycharmPlease = 69420
        assert isinstance(pycharmPlease, cls)
        return pycharmPlease
      else:
        pass

  def __eq__(self, other: Any) -> bool:
    """
    Check if the PathSyntaxException is equal to another object.

    Args:
      other (Any): The other object to compare.

    Returns:
      bool: True if equal, False otherwise.
    """
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    if self.badPath != other.badPath:
      return False
    return True



"""
Provides the example from the TypeSig documentation.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._type_sig_example import main

__all__ = [
    "main",
]



"""
ItemCall encapsulates a call to a namespace.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Union, Self, Callable, TypeAlias, Never


class ItemCall:
  """
  ItemCall encapsulates calls to the __setitem__, __getitem__,
  and __delitem__ methods of the AbstractNamespace class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  __slots__ = ('space', 'key', 'value', 'call')

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args) -> None:
    """
    Initializes the ItemCall instance with the provided arguments.

    Args:
      key: str
        The key for the item call.
      value: Any
        The value for the item call or the error or 'None' for deletion.
      call: FunctionType
        One of: 'dict.__setitem__', 'dict.__getitem__',
        or 'dict.__delitem__'.
    """
    if len(args) == 4:
      self.space, self.key, self.value, self.call = args
    elif len(args) == 3:
      self.space, self.key, self.call = args
      self.value = None
    else:
      infoSpec = """%s constructor expected 2 or 3 arguments, but received 
      %d!"""
      info = monoSpace(infoSpec % (type(self).__name__, len(args)))
      raise ValueError(info)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __hash__(self, ) -> int:
    """
    Returns the hash of the ItemCall instance.
    """
    try:
      if self.value is None:
        valueHash = 0
      else:
        valueHash = hash(self.value)
    except TypeError:
      valueHash = hash(object.__str__(self.value))
    callHashes = {
        dict.__setitem__: 1,
        dict.__getitem__: 2,
        dict.__delitem__: 3
    }
    for call, hash_ in callHashes.items():
      if self.call is call:
        break
    else:
      infoSpec = """Unable to resolve the call type for ItemCall: '%s'!"""
      info = monoSpace(infoSpec % (str(self.call),))
      raise ValueError(info)
    return hash((self.key, valueHash, hash_))

  def __eq__(self, other: Any) -> bool:
    """
    Compares the ItemCall instance with another object for equality.

    Args:
      other: Any
        The object to compare with.

    Returns:
      bool: True if the other object is an ItemCall and has the same key,
      value, and call, otherwise False.
    """
    if not isinstance(other, ItemCall):
      return False
    if self.space != other.space:
      return False
    if self.key != other.key:
      return False
    if self.value != other.value:
      return False
    if self.call is not other.call:
      return False
    return True

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def apply(self, space: dict) -> Any:
    """
    Applies the ItemCall to the provided namespace.

    Args:
      space: dict
        The namespace to apply the ItemCall to.

    Returns:
      dict: The updated namespace after applying the ItemCall.
    """
    if self.call is dict.__getitem__:
      return dict.__getitem__(space, self.key)
    if self.call is dict.__delitem__:
      try:
        dict.__delitem__(space, self.key)
      except Exception as exception:
        raise exception
      else:
        return space
      finally:
        pass  # NOQA
    if self.call is dict.__setitem__:
      dict.__setitem__(space, self.key, self.value)
      return space
    infoSpec = """%s expected 'call' to be one of: 'dict.__getitem__', 
    '__dict.__setitem__' or 'dict.__delitem__', but received '%s'!"""
    info = monoSpace(infoSpec % (type(self).__name__, str(self.call)))
    raise ValueError(info)



"""
The 'attributeErrorFactory' instantiates 'AttributeError' with a message
that matches the built-in message if a given object does not have a given
attribute. This allows simulating the absense of an attribute. For example:

class Bar:  # secret descriptor!

  __fallback_value__ = 'lol'
  __field_name__ = None
  __field_owner__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __get__(self, instance: object, owner: type) -> object:
    if instance is None:
      return self
    if hasattr(owner, '__trust_me_bro__'):
      pvtName = '__%s__' % self.__field_name__
      fallback = self.__fallback_value__
      return getattr(instance, pvtName, fallback)
    raise attributeErrorFactory(owner, self.__field_name__)


class Foo:
  __trust_me_bro__ = True
  bar = Bar()


class Sus:
  bar = Bar()


if __name__ == '__main__':
  foo = Foo()
  print(foo.bar)  # This will return the fallback value 'lol'
  sus = Sus()
  print(sus.bar)  # Raises AttributeError


The above attribute error will have the following message:
AttributeError: 'Sus' object has no attribute 'bar'
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class Breh(ValueError):
  """
  Breh is a custom exception that is raised when an attribute error is
  encountered, but the attribute error message does not match the built-in
  message.
  """

  posArgs = _Attribute()

  def __init__(self, *args: Any) -> None:
    self.posArgs = args
    ValueError.__init__(self, )

  def __str__(self) -> str:
    """
    Return a string representation of the Breh exception.
    """
    infoSpec = """Unable to resolve arguments: \n%s"""
    argStr = '<br><tab>'.join([str(arg) for arg in self.posArgs])
    info = infoSpec % argStr
    return monoSpace(info)

  __repr__ = __str__


def attributeErrorFactory(*args) -> AttributeError:
  """
  Factory function that creates an AttributeError with a message that
  matches the built-in message if a given object does not have a given
  attribute.
  """
  ownerName, fieldName = None, None
  for arg in args:
    if isinstance(arg, type) and ownerName is None:
      ownerName = arg.__name__
      continue
    if isinstance(arg, str):
      if ownerName is None:
        ownerName = arg
        continue
      if fieldName is None:
        fieldName = arg
        continue
  else:
    if ownerName is None or fieldName is None:
      raise Breh(*args)
  infoSpec = """AttributeError: '%s' object has no attribute '%s'"""
  info = infoSpec % (ownerName, fieldName)
  return AttributeError(info)  # type: ignore[return-value]



"""SubclassException should be raised when a class is not a subclass of
the expected base class. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any, Callable, Never


class SubclassException(TypeError):
  """SubclassException should be raised when a class is not a subclass of
  the expected base class."""

  actualObject = _Attribute(None)
  actualClass = _Attribute()
  expectedClass = _Attribute()

  def __init__(self, obj: object, expBase: type) -> None:
    """Initialize the exception with the object and expected base class."""
    if isinstance(obj, type):
      self.actualClass = obj
    else:
      self.actualClass = type(obj)
      self.actualObject = obj
    self.expectedClass = expBase
    if self.actualObject is not None:
      infoSpec = """Expected object of a subclass of '%s', but received 
      '%s' having mro: <br><tab>%s. """
      expName = expBase.__name__
      objStr = str(obj)
      mroList = [base for base in self.actualClass.__mro__]
      mroNames = [base.__name__ for base in mroList]
      mroStr = '<br><tab>'.join(mroNames)
      info = infoSpec % (expName, objStr, mroStr)
    else:
      infoSpec = """Expected subclass of '%s', but received '%s' having mro:
      <br><tab>%s. """
      expName = expBase.__name__
      objStr = ''
      mroList = [base for base in self.actualClass.__mro__]
      mroNames = [base.__name__ for base in mroList]
      mroStr = '<br><tab>'.join(mroNames)
      info = infoSpec % (expName, objStr, mroStr)
    TypeError.__init__(self, info)

  def _resolveOther(self, other: object) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the exception to another object."""
    cls = type(self)
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    if isinstance(other, cls):
      if self.actualObject != other.actualObject:
        return False
      if self.expectedClass != other.expectedClass:
        return False
      if self.actualClass != other.actualClass:
        return False
      return True
    return False



"""
The _KeeNumBase class provides a base for the KeeNum class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..attr import Field
from ..static import AbstractObject
from ..waitaminute import MissingVariable, TypeException
from ..waitaminute import IllegalInstantiation

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any

try:
  from icecream import ic  # NOQA

  ic.configureOutput(includeContext=True, )
except ImportError:
  print('icecream not installed, using dummy ic function.')
  ic = print


class _KeeNumBase(AbstractObject):
  """
  The _KeeNumBase class provides a base for the KeeNum class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __member_name__ = None  # Future name of the enumeration member
  __member_value__ = None  # Future value of the enumeration member
  __member_index__ = None  # Future index of the enumeration member
  __value_type__ = None  # Future type of the enumeration member values

  #  Public Variables
  name = Field()
  value = Field()
  index = Field()
  valueType = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  @name.GET
  def _getName(self) -> str:
    """Get the name of the enumeration member."""
    if self.__member_name__ is None:
      raise MissingVariable('__member_name__', str)
    if isinstance(self.__member_name__, str):
      return self.__member_name__
    raise TypeException('__member_name__', self.__member_name__, str)

  @value.GET
  def _getValue(self) -> Any:
    """Get the value of the enumeration member."""
    if self.__member_value__ is None:
      return self.__member_name__.upper()
    return self.__member_value__

  @index.GET
  def _getIndex(self) -> int:
    """Get the index of the enumeration member."""
    if self.__member_index__ is None:
      raise MissingVariable('__member_index__', int)
    if isinstance(self.__member_index__, int):
      return self.__member_index__
    raise TypeException('__member_index__', self.__member_index__, int)

  @valueType.GET
  def _getValueType(self) -> type:
    """Get the type of the enumeration member values."""
    if self.__value_type__ is None:
      raise MissingVariable('__value_type__', type)
    if isinstance(self.__value_type__, type):
      return self.__value_type__
    raise TypeException('__value_type__', self.__value_type__, type)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __str__(self) -> str:
    """
    String representation of the enumeration member.
    """
    clsName = type(self).__name__
    return """%s.%s""" % (clsName, self.name.upper())

  def __repr__(self) -> str:
    """
    Representation of the enumeration member.
    """
    typeName = self.valueType.__name__
    infoSpec = """<%s[%s]: %s>"""
    info = infoSpec % (str(self), typeName, repr(self.value))
    if len(info) > 77:
      return '%s...' % info[:74]
    return info

  def __hash__(self) -> int:
    """
    Hash representation of the enumeration member.
    """
    clsHash = hash(type(self))
    return hash((clsHash, self.name, self.index))

  def __eq__(self, other: Any) -> bool:
    """
    Equality check for the enumeration member.
    """
    if self is other:
      return True
    cls = type(self)
    if not isinstance(other, cls):
      return False
    return True if self.value == other.value else False

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args, **kwargs) -> None:
    """
    The control flow instantiates and initializes the members
    automatically. Do not reimplement or change this function. Developers
    of 'worktoy' considers this function guaranteed to remain unchanged.
    Changing it causes undefined behavior. For this reason, future
    versions are planned to specifically raise RuntimeError if this
    function is reimplemented or changed.

    As of development version 0.99.85, this function could be
    reimplemented by setting the '_root' keyword argument to True,
    but a much better method of preventing instantiation is in the works.
    This new prevention scheme will be backwards compatible, ensuring
    that libraries depending on 'worktoy.keenum' can safely upgrade,
    provided they do not reimplement this function!
    """
    if not kwargs.get('_root', False):
      raise IllegalInstantiation(type(self))

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #



"""AbstractException provides an abstract baseclass for custom exceptions
in the 'worktoy.waitaminute' module."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import logging
from abc import abstractmethod

from ..parse import maybe

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self, Never


class AbstractException(Exception):
  """AbstractException provides an abstract baseclass for custom exceptions
  in the 'worktoy.waitaminute' module."""

  def __new__(*_) -> Never:
    """
    I'll do it later, promise!
    """
    raise NotImplementedError



"""BadDelete is a temporary custom exception meant to be caught in
__delattr__. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class BadDelete(TypeError):
  """
  BadSet is a temporary custom class meant to be caught in __setattr__
  which raises a more precise error based on it.
  """

  instance = _Attribute()

  def __init__(self, instance: Any, ) -> None:
    """
    Initialize the BadSet instance.
    """
    self.instance = instance



"""
BaseSpace provides the namespace class used by worktoy.mcls.BaseMeta
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType as Func

from ..waitaminute import VariableNotNone
from ..static import TypeSig

from . import AbstractNamespace, AbstractMetaclass
from .hooks import OverloadHook, NameHook, ReservedNameHook, PreClassHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import TypeAlias, Callable, Any

  OverloadMap: TypeAlias = dict[str, dict[TypeSig, Callable[..., Any]]]


class BaseSpace(AbstractNamespace):
  """
  BaseSpace is the namespace used by BaseMeta. It enables function
  overloading and related features via hook registration.

  Classes defined using this namespace support method overloading through
  hooks installed automatically in the class body. These hooks handle
  overload collection, dispatch construction, and support for 'THIS' as a
  placeholder during class creation.

  The overload mechanism and other behavior are defined in
  `worktoy.mcls.hooks`. This namespace is returned from BaseMeta.__prepare__.
  """

  __overload_map__ = None

  def _buildOverloadMap(self, ) -> None:
    """Build the overload map for the namespace."""
    if self.__overload_map__ is not None:
      raise VariableNotNone('__overload_map__', )
    mcls = self.getMetaclass()
    bases = self.getBases()
    baseSpaces = []
    self.__overload_map__ = {}
    for base in bases:
      if isinstance(base, AbstractMetaclass):
        baseSpaces.append(base.getNamespace())
    while baseSpaces:
      space = baseSpaces.pop()
      baseOverloadMap = space.getOverloadMap()
      for key, overloads in baseOverloadMap.items():
        existing = self.__overload_map__.get(key, {})
        for sig, func in overloads.items():
          existing[sig] = func
        self.__overload_map__[key] = existing

  def getOverloadMap(self, **kwargs) -> OverloadMap:
    """Get the overload map for the namespace."""
    if self.__overload_map__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self._buildOverloadMap()
      return self.getOverloadMap(_recursion=True)
    return self.__overload_map__

  def addOverload(self, key: str, sig: TypeSig, func: Func) -> None:
    """Add an overload to the overload map."""
    overloadMap = self.getOverloadMap()
    existingMap = overloadMap.get(key, {})
    existingMap[sig] = func
    overloadMap[key] = existingMap
    self.__overload_map__ = overloadMap

  reservedNameHook = ReservedNameHook()
  preClassHook = PreClassHook()
  overloadHook = OverloadHook()
  nameHook = NameHook()
